# -*- coding: utf8 -*-
import re
import socket
import errno
import json
import struct
import traceback

from six import iteritems
from werkzeug.datastructures import ImmutableMultiDict as _ImmutableMultiDict

from .interface import BUIbackend
from ..parser.interface import BUIparser
from ...exceptions import BUIserverException
from ..._compat import pickle, to_unicode, to_bytes
from ...decorators import implement
from ...datastructures import ImmutableMultiDict


INTERFACE_METHODS = BUIbackend.__abstractmethods__
PARSER_INTERFACE_METHODS = BUIparser.__abstractmethods__
AGENT_VERSION_CAST = '0.4.9999'


class ProxyCall(object):
    """Class to dispatch call of unknown methods in order to dynamically
    call agents one without maintaining the explicit list of methods.
    """
    def __init__(self, proxy, method, network=False):
        """
        :param proxy: Class to proxify
        :type proxy: :class:`burpui.misc.backend.multi.Burp`

        :param method: Name of the method to proxify
        :type method: str

        :param network: Is it a custom call over network
        :type network: bool
        """
        self.proxy = proxy
        self.method = method
        self.network = network

    def __call__(self, *args, **kwargs):
        """This is where the proxy call (and the magic) occurs"""
        # retrieve the original function prototype
        proto = getattr(BUIbackend, self.method)
        args_name = list(proto.__code__.co_varnames)
        # skip self
        args_name.pop(0)
        # we transform unnamed arguments to named ones
        # example:
        #     def my_function(toto, tata=None, titi=None):
        #
        #     x = my_function('blah', titi='blih')
        #
        # => {'toto': 'blah', 'titi': 'blih'}
        encoded_args = {}
        for idx, opt in enumerate(args):
            encoded_args[args_name[idx]] = opt
        encoded_args.update(kwargs)

        # Special case for network calls
        if self.network:
            data = {'func': self.method, 'args': encoded_args}
            return json.loads(self.proxy.do_command(data))
        # normal case for "standard" interface
        if 'agent' not in encoded_args:
            raise AttributeError(str(encoded_args))
        agentName = encoded_args['agent']
        # we don't need this argument anymore
        del encoded_args['agent']
        try:
            agent = self.proxy.servers[agentName]
        except KeyError:
            # This exception should be forwarded to the final user
            if not agentName:
                msg = "You must provide an agent name"
            else:
                msg = "Agent '{}' not found".format(agentName)
            raise BUIserverException(msg)
        return getattr(agent, self.method)(**encoded_args)


class ProxyParserCall(object):
    """Class that actually calls the Parser method"""
    def __init__(self, agent, method):
        """
        :param agent: Agent to use
        :type agent: :class:`burpui.misc.backend.multi.NClient`

        :param method: Name of the method to proxify
        :type method: str
        """
        self.agent = agent
        self.method = method

    def __call__(self, *args, **kwargs):
        """This is where the proxy call (and the magic) occurs"""
        # retrieve the original function prototype
        proto = getattr(BUIparser, self.method)
        args_name = list(proto.__code__.co_varnames)
        # skip self
        args_name.pop(0)
        # we transform unnamed arguments to named ones
        # example:
        #     def my_function(toto, tata=None, titi=None):
        #
        #     x = my_function('blah', titi='blih')
        #
        # => {'toto': 'blah', 'titi': 'blih'}
        encoded_args = {}
        for idx, opt in enumerate(args):
            encoded_args[args_name[idx]] = opt
        encoded_args.update(kwargs)

        data = {'func': 'proxy_parser', 'method': self.method, 'args': encoded_args}
        return json.loads(self.agent.do_command(data))


class ProxyParser(BUIparser):
    """Class to generate a "virtual" parser object"""
    # These functions MUST be implemented because we inherit an abstract class.
    # The hack here is to get the list of the functions and let the interpreter
    # think we don't have to implement them.
    # Thanks to this list, we know what function are implemented by our backend.
    foreign = PARSER_INTERFACE_METHODS
    BUIparser.__abstractmethods__ = frozenset()

    def __init__(self, agent):
        """
        :param agent: Agent to use
        :type agent: :class:`burpui.misc.backend.multi.NClient`
        """
        self.agent = agent

    def __getattribute__(self, name):
        # always return this value because we need it and if we don't do that
        # we'll end up with an infinite loop
        if name == 'foreign':
            return object.__getattribute__(self, name)
        # now we can retrieve the 'foreign' list and know if the object called
        # needs to be "proxyfied"
        if name in self.foreign:
            return ProxyParserCall(self.agent, name)
        return object.__getattribute__(self, name)


class Burp(BUIbackend):
    """The :class:`burpui.misc.backend.multi.Burp` class provides a consistent
    backend to interact with ``agents``.

    It is actually the *real* multi backend implementing the
    :class:`burpui.misc.backend.interface.BUIbackend` class.

    For each agent found in the configuration, it will load a
    :class:`burpui.misc.backend.multi.NClient` class.

    :param server: ``Burp-UI`` server instance in order to access logger
                   and/or some global settings
    :type server: :class:`burpui.server.BUIServer`

    :param conf: Configuration file to use
    :type conf: str
    """
    # These functions MUST be implemented because we inherit an abstract class.
    # The hack here is to get the list of the functions and let the interpreter
    # think we don't have to implement them.
    # Thanks to this list, we know what function are implemented by our backend.
    foreign = INTERFACE_METHODS
    BUIbackend.__abstractmethods__ = frozenset()

    def __init__(self, server=None, conf=None):
        """
        :param server: Application context
        :type server: :class:`burpui.server.BUIServer`

        :param conf: Configuration
        :type conf: :class:`burpui.config.BUIConfig`
        """
        self.app = server
        self.acl_handler = server.acl_handler
        self.servers = {}
        self.app.config['SERVERS'] = []
        if conf:
            for sect in conf.options.keys():
                r = re.match('^Agent:(.+)$', sect)
                if r:
                    host = conf.safe_get('host', section=sect)
                    port = conf.safe_get('port', 'integer', section=sect) or 10000
                    password = conf.safe_get('password', section=sect)
                    ssl = conf.safe_get('ssl', 'boolean', section=sect) or False
                    timeout = conf.safe_get('timeout', 'integer', section=sect) or 5

                    self.servers[r.group(1)] = NClient(self.app, host, port, password, ssl, timeout)
                    self.app.config['SERVERS'].append(r.group(1))

        if not self.servers:
            self.logger.error('No agent configured!')
        else:
            self.logger.debug(self.servers)

    def __getattribute__(self, name):
        # always return this value because we need it and if we don't do that
        # we'll end up with an infinite loop
        if name == 'foreign':
            return object.__getattribute__(self, name)
        # now we can retrieve the 'foreign' list and know if the object called
        # needs to be "proxyfied"
        if name in self.foreign:
            proxy = True
            func = None
            try:
                func = object.__getattribute__(self, name)
                proxy = not getattr(func, '__ismethodimplemented__', False)
            except:
                pass
            self.logger.debug('func: {} - {}'.format(name, proxy))
            if proxy:
                return ProxyCall(self, name)
            elif func:
                return func
        return object.__getattribute__(self, name)

    @implement
    def is_one_backup_running(self, agent=None):
        """See :func:`burpui.misc.backend.interface.BUIbackend.is_one_backup_running`"""
        r = []
        if agent:
            try:
                r = self.servers[agent].is_one_backup_running(agent)
            except BUIserverException:
                pass
        else:
            r = {}
            for name, serv in iteritems(self.servers):
                try:
                    r[name] = serv.is_one_backup_running()
                except BUIserverException:
                    r[name] = []
        return r

    def _get_version(self, method=None):
        """get versions"""

        if not method:
            raise BUIserverException('Wrong method call')

        r = {}

        for name, serv in iteritems(self.servers):
            func = getattr(serv, method)
            try:
                r[name] = func()
            except BUIserverException:
                r[name] = 'Unknown'

        return r

    @implement
    def get_parser(self, agent=None):
        # Need to return a proxy object to interact with a remote parser
        if not agent:
            raise BUIserverException('No agent provided')

        return ProxyParser(self.servers.get(agent))

    @implement
    def get_client_version(self, agent=None):
        """See :func:`burpui.misc.backend.interface.BUIbackend.get_client_version`"""
        if not agent:
            return self._get_version('get_client_version')
        return self.servers[agent].get_client_version()

    @implement
    def get_server_version(self, agent=None):
        """See :func:`burpui.misc.backend.interface.BUIbackend.get_server_version`"""
        if not agent:
            return self._get_version('get_server_version')
        return self.servers[agent].get_server_version()


class Gsocket():
    def __init__(self, host, port, ssl=False, timeout=5, notimeout=False):
        self.host = host
        self.port = port
        self.ssl = ssl
        self.timeout = timeout
        self.notimeout = notimeout

    def conn(self):
        if self.ssl:
            import ssl
            s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            if not self.notimeout:
                s.settimeout(self.timeout)
            s.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
            ret = ssl.wrap_socket(s, cert_reqs=ssl.CERT_NONE, ssl_version=ssl.PROTOCOL_SSLv23)
            ret.connect((self.host, self.port))
        else:
            if not self.notimeout:
                ret = socket.create_connection((self.host, self.port), timeout=self.timeout)
            else:
                ret = socket.create_connection((self.host, self.port))
            ret.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
        self.sock = ret
        self.connected = True

    def __enter__(self):
        self.conn()
        return self.sock, self

    def __exit__(self, type, value, traceback):
        if self.connected:
            self.sock.close()
        self.connected = False

    def recvall(self, length=1024):
        """Read the answer of the agent"""
        buf = b''
        bsize = 1024
        received = 0
        if length < bsize:
            bsize = length
        while received < length:
            newbuf = self.sock.recv(bsize)
            if not newbuf:
                return None
            buf += newbuf
            received += len(newbuf)
        return buf


class NClient(BUIbackend):
    """The :class:`burpui.misc.backend.multi.NClient` class provides a
    consistent backend to interact with ``agents``.

    It acts as a proxy so it works with any agent running a backend implementing
    the :class:`burpui.misc.backend.interface.BUIbackend` class.

    :param app: The application context
    :type app: :class:`burpui.server.BUIServer`

    :param host: Address of the remote agent
    :type host: str

    :param port: Port of the remote agent
    :type port: int

    :param password: Secret between the agent and the burp-ui server
    :type password: str

    :param ssl: Use SSL to communicate with the agent
    :type ssl: bool
    """
    # These functions MUST be implemented because we inherit an abstract class.
    # The hack here is to get the list of the functions and let the interpreter
    # think we don't have to implement them.
    # Thanks to this list, we know what function are implemented by our backend.
    foreign = INTERFACE_METHODS
    BUIbackend.__abstractmethods__ = frozenset()

    def __init__(self, app=None, host=None, port=None, password=None, ssl=None, timeout=5):
        self.host = host
        self.port = port
        self.password = password
        self.ssl = ssl
        self.app = app
        self.timeout = timeout or 5
        self._agent_version = None

    def __getattribute__(self, name):
        # always return this value because we need it and if we don't do that
        # we'll end up with an infinite loop
        if name == 'foreign':
            return object.__getattribute__(self, name)
        # now we can retrieve the 'foreign' list and know if the object called
        # needs a dynamic implementation
        if name in self.foreign:
            proxy = True
            func = None
            try:
                func = object.__getattribute__(self, name)
                proxy = not getattr(func, '__ismethodimplemented__', False)
            except:
                pass
            self.logger.debug('func: {} - {}'.format(name, proxy))
            if proxy:
                return ProxyCall(self, name, network=True)
            elif func:
                return func
        return object.__getattribute__(self, name)

    def _get_agent_version(self):
        if self.ping() and not self._agent_version:
            data = {'func': 'agent_version'}
            try:
                vers = self.do_command(data)
                self._agent_version = json.loads(to_unicode(vers))
            except BUIserverException:
                # just ignore the error if this custom function is not
                # implemented
                pass
        return self._agent_version

    def ping(self):
        """Check if we are connected to the agent"""
        res = False
        try:
            with Gsocket(self.host, self.port, self.ssl, self.timeout) as (sock, gsock):
                sock.sendall(struct.pack('!Q', 2))
                sock.sendall(b'RE')
                res = True
        except socket.error:
            pass
        return res

    def setup(self, sock, gsock, data):
        length = struct.pack('!Q', len(data))
        sock.sendall(length)
        data = to_unicode(data)
        sock.sendall(to_bytes(data))
        self.logger.debug("Sending: {}".format(data))
        tmp = to_unicode(sock.recv(2))
        self.logger.debug("recv: '{}'".format(tmp))
        if 'ER' == tmp:
            lengthbuf = sock.recv(8)
            length, = struct.unpack('!Q', lengthbuf)
            err = to_unicode(gsock.recvall(length))
            raise BUIserverException(err)
        if 'OK' != tmp:
            self.logger.debug('Ooops, unsuccessful!')
            return False
        self.logger.debug("Data sent successfully")
        return True

    def do_command(self, data=None, restarted=False):
        """Send a command to the remote agent"""
        res = '[]'
        err = None
        notimeout = False
        timeout = self.timeout
        if not data:
            raise BUIserverException('Missing data')
        data['password'] = self.password
        # manage long running operations
        if data['func'] in ['restore_files', 'get_file', 'del_file']:
            notimeout = True
        if data['func'] == 'get_tree' and data['args'].get('root') == '*':
            # arbitrary raise timeout
            timeout = max(timeout, 300)
        try:
            # don't need a context manager here
            if data['func'] == 'get_file':
                gsock = Gsocket(self.host, self.port, self.ssl, notimeout=True)
                gsock.conn()
                raw = json.dumps(data)
                if not self.setup(gsock.sock, gsock, raw):
                    return res
                return gsock.sock
            with Gsocket(self.host, self.port, self.ssl, timeout, notimeout) as (sock, gsock):
                try:
                    raw = json.dumps(data)
                    if not self.setup(gsock.sock, gsock, raw):
                        return res
                    lengthbuf = sock.recv(8)
                    length, = struct.unpack('!Q', lengthbuf)
                    res = to_unicode(gsock.recvall(length))
                except IOError as e:
                    if not restarted and e.errno == errno.EPIPE:
                        self.logger.warning('Broken pipe, restarting the request')
                        return self.do_command(data, True)
                    elif e.errno == errno.ECONNRESET:
                        self.logger.error('!!! {} !!!\nPlease check your SSL configuration on both sides!'.format(str(e)))
                    else:
                        self.logger.error('!!! {} !!!\n{}'.format(str(e), traceback.format_exc()))
                    raise e
                except socket.timeout as e:
                    if self.app.gunicorn and not restarted:
                        self.logger.warning('Socket timed-out, restarting the request')
                        return self.do_command(data, True)
                    self.logger.error('!!! {} !!!\n{}'.format(str(e), traceback.format_exc()))
                    raise e
                # catch all
                except Exception as e:
                    self.logger.error('!!! {} !!!\n{}'.format(str(e), traceback.format_exc()))
                    if data['func'] == 'restore_files':
                        err = str(e)
                    elif isinstance(e, BUIserverException):
                        raise e
                    else:
                        raise BUIserverException(str(e))
        except Exception as e:
            self.logger.error('!!! {} !!!\n{}'.format(str(e), traceback.format_exc()))
            raise BUIserverException(str(e))

        if data['func'] == 'restore_files':
            if err:
                res = None
            return res, err

        return res

    """
    Utilities functions
    """

    @implement
    def store_conf_cli(self, data, client=None, conf=None, template=False, agent=None):
        """See :func:`burpui.misc.backend.interface.BUIbackend.store_conf_cli`"""
        # serialize data as it is a nested dict
        import hmac
        import hashlib
        from base64 import b64encode
        if not isinstance(data, (_ImmutableMultiDict, ImmutableMultiDict)):
            msg = 'Wrong data type'
            self.logger.warning(msg)
            raise BUIserverException(msg)
        vers = self._get_agent_version()
        if vers and vers >= AGENT_VERSION_CAST:
            # convert the data to our custom ImmutableMultiDict
            data = ImmutableMultiDict(data.to_dict(False))
        key = '{}{}'.format(self.password, 'store_conf_cli')
        key = to_bytes(key)
        pickles = to_unicode(b64encode(pickle.dumps({'data': data, 'conf': conf, 'client': client, 'template': template}, 2)))
        bytes_pickles = to_bytes(pickles)
        digest = to_unicode(hmac.new(key, bytes_pickles, hashlib.sha1).hexdigest())
        data = {'func': 'store_conf_cli', 'args': pickles, 'pickled': True, 'digest': digest}
        return json.loads(self.do_command(data))

    @implement
    def store_conf_srv(self, data, conf=None, agent=None):
        """See :func:`burpui.misc.backend.interface.BUIbackend.store_conf_srv`"""
        # serialize data as it is a nested dict
        import hmac
        import hashlib
        from base64 import b64encode
        if not isinstance(data, (_ImmutableMultiDict, ImmutableMultiDict)):
            msg = 'Wrong data type'
            self.logger.warning(msg)
            raise BUIserverException(msg)
        vers = self._get_agent_version()
        if vers and vers >= AGENT_VERSION_CAST:
            # convert the data to our custom ImmutableMultiDict
            data = ImmutableMultiDict(data.to_dict(False))
        key = u'{}{}'.format(self.password, 'store_conf_srv')
        key = to_bytes(key)
        pickles = to_unicode(b64encode(pickle.dumps({'data': data, 'conf': conf}, 2)))
        bytes_pickles = to_bytes(pickles)
        digest = to_unicode(hmac.new(key, bytes_pickles, hashlib.sha1).hexdigest())
        data = {'func': 'store_conf_srv', 'args': pickles, 'pickled': True, 'digest': digest}
        return json.loads(self.do_command(data))

    @implement
    def restore_files(self, name=None, backup=None, files=None, strip=None, archive='zip', password=None, agent=None):
        """See :func:`burpui.misc.backend.interface.BUIbackend.restore_files`"""
        data = {'func': 'restore_files', 'args': {'name': name, 'backup': backup, 'files': files, 'strip': strip, 'archive': archive, 'password': password}}
        return self.do_command(data)

    @implement
    def get_file(self, path, agent=None):
        """See :func:`burpui.misc.backend.interface.BUIbackend.get_file`"""
        data = {'func': 'get_file', 'path': path}
        return self.do_command(data)

    @implement
    def del_file(self, path, agent=None):
        """See :func:`burpui.misc.backend.interface.BUIbackend.del_file`"""
        data = {'func': 'del_file', 'path': path}
        return json.loads(self.do_command(data))



# -*- coding: utf8 -*-
"""
.. module:: burpui.api
    :platform: Unix
    :synopsis: Burp-UI api module.

.. moduleauthor:: Ziirish <hi+burpui@ziirish.me>


"""
import os
import sys
import uuid
import hashlib
import logging

from flask import Blueprint, Response, request, current_app, session, abort
from flask_restplus import Api as ApiPlus
from flask_login import current_user
from importlib import import_module
from functools import wraps

from .custom.namespace import Namespace
from .._compat import to_bytes
from ..desc import __version__, __release__, __url__, __doc__
from ..server import BUIServer  # noqa
from ..exceptions import BUIserverException
from ..config import config

bui = current_app  # type: BUIServer
EXEMPT_METHODS = set(['OPTIONS'])


def force_refresh():
    return request.headers.get('X-No-Cache', False) is not False


def cache_key():
    key = '{}-{}-{}-{}-{}-{}'.format(
        session.get('login', uuid.uuid4()),
        request.path,
        request.values,
        request.headers.get('X-Session-Tag', ''),
        request.cookies,
        session.get('language', '')
    )
    key = hashlib.sha256(to_bytes(key)).hexdigest()
    return key


def api_login_required(func):
    """Custom login decorator that is able to parse Basic credentials as well as
    Cookies set with the traditional login.
    """
    @wraps(func)
    def decorated_view(*args, **kwargs):
        """decorator"""
        if request.method in EXEMPT_METHODS:  # pragma: no cover
            return func(*args, **kwargs)
        # 'func' is a Flask.view.MethodView so we have access to some special
        # params
        cls = func.view_class
        login_required = getattr(cls, 'login_required', True)
        if (bui.auth != 'none' and
                login_required and
                not bui.config.get('LOGIN_DISABLED', False)):
            if not current_user.is_authenticated:
                if request.headers.get('X-From-UI', False):
                    abort(403)
                return Response(
                    'Could not verify your access level for that URL.\n'
                    'You have to login with proper credentials', 401,
                    {'WWW-Authenticate': 'Basic realm="Login Required"'})
        return func(*args, **kwargs)
    return decorated_view


def check_acl(func):
    """Custom decorator to check if the ACL are in use or not"""
    @wraps(func)
    def decorated_view(*args, **kwargs):
        if request.method in EXEMPT_METHODS:  # pragma: no cover
            return func(*args, **kwargs)
        # 'func' is a Flask.view.MethodView so we have access to some special
        # params
        cls = func.view_class
        login_required = getattr(cls, 'login_required', True)
        if (bui.auth != 'none' and
                login_required and
                not bui.config.get('LOGIN_DISABLED', False)):
            if current_user.is_anonymous:
                abort(403)
        return func(*args, **kwargs)
    return decorated_view


class Api(ApiPlus):
    """Wrapper class around :class:`flask_restplus.Api`"""
    logger = logging.getLogger('burp-ui')
    # TODO: should use global object instead of reference
    loaded = False
    release = __release__
    __doc__ = __doc__
    __url__ = __url__
    CELERY_REQUIRED = ['async']

    def load_all(self):
        if config['WITH_LIMIT']:
            try:
                from ..ext.limit import limiter
                self.decorators.append(limiter.limit(config['BUI_RATIO']))
            except ImportError:
                self.logger.warning('Unable to import limiter module')
        """hack to automatically import api modules"""
        if not self.loaded:
            sys.path.insert(0, os.path.dirname(os.path.abspath(__file__)))
            self.loaded = True
            for f in os.listdir(__path__[0]):
                name, ext = os.path.splitext(f)
                if (os.path.isfile(os.path.join(__path__[0], f)) and
                        ext == '.py' and
                        name not in ['__init__', '.', '..']):
                    mod = '.' + name
                    if name not in self.CELERY_REQUIRED or \
                            config['WITH_CELERY']:
                        self.logger.debug('Loading API module: {}'.format(mod))
                        try:
                            import_module(mod, __name__)
                        except:  # pragma: no cover
                            import traceback
                            self.logger.critical(
                                'Unable to load {}:\n{}'.format(
                                    mod,
                                    traceback.format_exc()
                                )
                            )
                    else:
                        self.logger.warning(
                            'Skipping API module: {}'.format(mod)
                        )

    def acl_admin_required(self, message='Access denied', code=403):
        def decorator(func):
            @wraps(func)
            def decorated(resource, *args, **kwargs):
                if not current_user.is_anonymous and \
                        not current_user.acl.is_admin():
                    resource.abort(code, message)
                return func(resource, *args, **kwargs)
            return decorated
        return decorator

    def acl_admin_or_moderator_required(self, message='Access denied', code=403):
        def decorator(func):
            @wraps(func)
            def decorated(resource, *args, **kwargs):
                if not current_user.is_anonymous and \
                        not current_user.acl.is_admin() and \
                        not current_user.acl.is_moderator():
                    resource.abort(code, message)
                return func(resource, *args, **kwargs)
            return decorated
        return decorator

    def acl_own_or_admin(self, key='name', message='Access denied', code=403):
        def decorator(func):
            @wraps(func)
            def decorated(resource, *args, **kwargs):
                if key not in kwargs:  # pragma: no cover
                    resource.abort(500, "key '{}' not found".format(key))
                if kwargs[key] != current_user.name and \
                        not current_user.is_anonymous and \
                        not current_user.acl.is_admin():
                    resource.abort(code, message)
                return func(resource, *args, **kwargs)
            return decorated
        return decorator

    def disabled_on_demo(self):
        def decorator(func):
            @wraps(func)
            def decorated(resource, *args, **kwargs):
                if config['BUI_DEMO']:
                    resource.abort(
                        405,
                        'Sorry, this feature is not available on the demo'
                    )
                return func(resource, *args, **kwargs)
            return decorated
        return decorator

    def namespace(self, *args, **kwargs):
        """A namespace factory

        :returns Namespace: a new namespace instance
        """
        ns = Namespace(*args, **kwargs)
        self.add_namespace(ns)
        return ns


apibp = Blueprint('api', __name__, url_prefix='/api')
api = Api(
    apibp,
    title='Burp-UI API',
    description='Burp-UI API to interact with burp',
    version=__version__,
    doc='/doc',
    decorators=[api_login_required]
)


@api.errorhandler(BUIserverException)
def handle_bui_server_exception(error):
    """Forward a BUIserverException to the final user

    :param error: Custom exception
    :type error: :class:`burpui.exceptions.BUIserverException`
    """
    return {'message': error.description}, error.code



"""
Burp-UI is a web-ui for burp backup written in python with Flask and
jQuery/Bootstrap

.. module:: burpui.tasks
    :platform: Unix
    :synopsis: Burp-UI tasks module.

.. moduleauthor:: Ziirish <hi+burpui@ziirish.me>
"""
import sys
import os

from six import iteritems
from flask import current_app
from datetime import timedelta, datetime
from celery.schedules import crontab
from celery.utils.log import get_task_logger
from time import gmtime, strftime, sleep

# Try to load modules from our current env first
sys.path.insert(0, os.path.join(os.path.dirname(os.path.realpath(__file__)), '..'))

from burpui._compat import PY3  # noqa
from burpui.config import config  # noqa
from burpui.ext.async import celery  # noqa
from burpui.ext.cache import cache  # noqa
from burpui.sessions import session_manager  # noqa
from burpui.server import BUIServer  # noqa
from burpui.exceptions import BUIserverException  # noqa
from burpui.api.client import ClientTreeAll  # noqa

try:
    from .ext.ws import socketio
    WS_AVAILABLE = True
except ImportError:
    WS_AVAILABLE = False

if not PY3:
    from itertools import imap as map

if config.get('WITH_SQL'):
    from burpui.ext.sql import db
else:
    db = None

bui = current_app  # type: BUIServer
logger = get_task_logger(__name__)
ME = __name__

LOCK_EXPIRE = 60 * 30  # Lock expires in 30 minutes

BEAT_SCHEDULE = {
    'ping-backend-hourly': {
        'task': '{}.ping_backend'.format(ME),
        'schedule': crontab(minute='15'),  # run every hour
    },
    'backup-running-4-minutely': {
        'task': '{}.backup_running'.format(ME),
        'schedule': timedelta(seconds=15),  # run every 15 seconds
    },
    'get-all-backups-every-twenty-minutes': {
        'task': '{}.get_all_backups'.format(ME),
        'schedule': crontab(minute='*/20'),  # every 20 minutes
    },
    'get-all-clients-reports-every-twenty-minutes': {
        'task': '{}.get_all_clients_reports'.format(ME),
        'schedule': crontab(minute='*/20'),  # every 20 minutes
    },
    'cleanup-expired-sessions-every-four-hours': {
        'task': '{}.cleanup_expired_sessions'.format(ME),
        'schedule': crontab(hour='*/4'),  # every four hours
    },
}

if db:
    from burpui.models import Task

    BEAT_SCHEDULE.update({
        'cleanup-restore-hourly': {
            'task': '{}.cleanup_restore'.format(ME),
            'schedule': crontab(minute='12'),  # run every hour
        },
    })

if 'CELERYBEAT_SCHEDULE' in celery.conf and \
        isinstance(celery.conf['CELERYBEAT_SCHEDULE'], dict):
    celery.conf['CELERYBEAT_SCHEDULE'].update(BEAT_SCHEDULE)
else:
    celery.conf['CELERYBEAT_SCHEDULE'] = BEAT_SCHEDULE


def acquire_lock(name, value='nyan', timeout=LOCK_EXPIRE):
    """Utility function to acquire a lock before processing the request"""
    lock = cache.cache.get(name)
    if lock:
        acquire_lock.lock = lock
        return False
    return cache.cache.add(name, value, timeout)


acquire_lock.lock = None


def release_lock(name):
    """Utility function to release a lock"""
    return cache.cache.delete(name)


def wait_for(lock_name, value, wait=10, timeout=LOCK_EXPIRE):
    """Utility function to wait until the given lock has been released"""
    old_lock = None
    if not acquire_lock(lock_name, value, timeout):
        logger.warn(
            'A task is already running. Wait for it: {}/{}'.format(
                lock_name,
                acquire_lock.lock
            )
        )
        old_lock = acquire_lock.lock
        # The lock should be released after LOCK_EXPLIRE max
        while not acquire_lock(lock_name, value, timeout):
            sleep(wait)

        # TODO: maybe we should check the status of task referenced by"old_lock"
        # to make sure the lock did not expire and the task is actually over
        logger.debug('lock released by: {}'.format(old_lock))

    return old_lock


@celery.task(ignore_result=True)
def ping_backend():
    if bui.standalone:
        bui.client.status()
    else:
        def __status(server):
            (serv, back) = server
            try:
                return bui.client.status(agent=serv)
            except BUIserverException:
                return False

        list(map(
            __status,
            iteritems(bui.client.servers)
        ))


@celery.task(bind=True, ignore_result=True)
def backup_running(self):
    # run one at the time, if one task was already running, we just discard
    # the new attempt
    if not acquire_lock(self.name):
        return None
    try:
        res = bui.client.is_one_backup_running()
        cache.cache.set(
            'backup_running_result',
            res,
            120
        )
        if WS_AVAILABLE:
            running = False
            if isinstance(res, dict):
                for _, run in iteritems(res):
                    if len(run) > 0:
                        running = True
                        break
            elif len(res) > 0:
                running = True
            socketio.emit('backup_running', running, namespace='/ws')
    finally:
        release_lock(self.name)


@celery.task(bind=True, ignore_result=True)
def get_all_backups(self):
    # run one at the time, if one task was already running, we just discard
    # the new attempt
    if not acquire_lock(self.name):
        return None
    try:
        backups = {}
        if bui.standalone:
            for cli in bui.client.get_all_clients():
                backups[cli['name']] = bui.client.get_client(cli['name'])
        else:
            for serv in bui.client.servers:
                backups[serv] = {}
                for cli in bui.client.get_all_clients(agent=serv):
                    backups[serv][cli['name']] = bui.client.get_client(cli['name'], agent=serv)
        cache.cache.set('all_backups', backups, 3600)
    finally:
        release_lock(self.name)


@celery.task(bind=True, ignore_result=True)
def get_all_clients_reports(self):
    # run one at the time, if one task was already running, we just discard
    # the new attempt
    if not acquire_lock(self.name):
        return None
    try:
        reports = {}
        if bui.standalone:
            reports = bui.client.get_clients_report(bui.client.get_all_clients())
        else:
            for serv in bui.client.servers:
                reports[serv] = bui.client.get_clients_report(bui.client.get_all_clients(agent=serv), serv)
        cache.cache.set('all_clients_reports', reports, 3600)
    finally:
        release_lock(self.name)


@celery.task(ignore_result=True)
def cleanup_expired_sessions():
    def expires(sess):
        ret = session_manager.invalidate_session_by_id(sess.uuid)
        if ret:
            session_manager.delete_session_by_id(sess.uuid)
        return ret
    list(map(expires, session_manager.get_expired_sessions()))


@celery.task(ignore_result=True)
def cleanup_restore():
    tasks = db.session.query(Task).filter(Task.task == 'perform_restore').filter(datetime.utcnow() > Task.expire).all()
    for rec in tasks:
        logger.info('Task expired: {}'.format(rec))
        task = perform_restore.AsyncResult(rec.uuid)
        try:
            if task.state != 'SUCCESS':
                logger.warn(
                    'Task is not done yet or did not end '
                    'successfully: {}'.format(task.state)
                )
                continue
            if not task.result:
                logger.warn('The task did not return anything')
                continue
            server = task.result.get('server')
            path = task.result.get('path')
            if path:
                if server:
                    if not bui.client.del_file(path, agent=server):
                        logger.warn("'{}' already removed".format(path))
                else:
                    if os.path.isfile(path):
                        os.unlink(path)
        finally:
            try:
                db.session.delete(rec)
                db.session.commit()
            except:
                db.session.rollback()
            task.revoke()


@celery.task(bind=True)
def perform_restore(self, client, backup,
                    files, strip, fmt, passwd, server=None, user=None,
                    admin=False, room=None, expire=timedelta(minutes=60)):
    ret = None
    lock_name = '{}-{}'.format(self.name, server)

    # TODO: maybe do something with old_lock someday
    wait_for(lock_name, self.request.id)

    try:
        if server:
            filename = 'restoration_%d_%s_on_%s_at_%s.%s' % (
                backup,
                client,
                server,
                strftime("%Y-%m-%d_%H_%M_%S", gmtime()),
                fmt)
        else:
            filename = 'restoration_%d_%s_at_%s.%s' % (
                backup,
                client,
                strftime("%Y-%m-%d_%H_%M_%S", gmtime()),
                fmt)

        self.update_state(state='STARTED', meta={'step': 'doing'})
        archive, err = bui.client.restore_files(
            client,
            backup,
            files,
            strip,
            fmt,
            passwd,
            server
        )
        if not archive:
            if not err:
                err = 'Something went wrong while restoring'
            self.update_state(state='FAILURE', meta={'error': err})
            logger.error('FAILURE: {}'.format(err))
        else:
            ret = {
                'filename': filename,
                'path': archive,
                'user': user,
                'server': server,
                'admin': admin
            }
            logger.debug(ret)

    finally:
        release_lock(lock_name)

    if db:
        curr = Task.query.filter_by(uuid=self.request.id).first()
        if curr:
            curr.expire = datetime.utcnow() + expire
            try:
                db.session.commit()
            except:
                db.session.rollback()

    if err:
        # make the task crash
        raise Exception(err)

    return ret


@celery.task(bind=True)
def load_all_tree(self, client, backup, server=None, user=None):
    key = 'load_all_tree-{}-{}-{}'.format(client, backup, server)
    ret = cache.cache.get(key)
    if ret:
        return {
            'client': client,
            'backup': backup,
            'server': server,
            'user': user,
            'tree': ret
        }

    lock_name = '{}-{}'.format(self.name, server)

    # TODO: maybe do something with old_lock someday
    wait_for(lock_name, self.request.id)

    try:
        ret = ClientTreeAll._get_tree_all(client, backup, server)
    except BUIserverException as exp:
        raise Exception(str(exp))
    finally:
        release_lock(lock_name)

    cache.cache.set(key, ret, 3600)
    return {
        'client': client,
        'backup': backup,
        'server': server,
        'user': user,
        'tree': ret
    }


def force_scheduling_now():
    """Force scheduling some tasks now"""
    get_all_backups.delay()
    backup_running.delay()
    get_all_clients_reports.delay()
    cleanup_expired_sessions.delay()
    cleanup_restore.delay()



# -*- coding: utf8 -*-
"""
.. module:: burpui.ext.ws
    :platform: Unix
    :synopsis: Burp-UI external WebSocket module.

.. moduleauthor:: Ziirish <hi+burpui@ziirish.me>

"""
from ..config import config
from flask_socketio import SocketIO

options = {}
if config.get('WS_ASYNC_MODE'):
    options['async_mode'] = config.get('WS_ASYNC_MODE')

#    engineio_logger=config.get('WS_DEBUG', False),
socketio = SocketIO(
    message_queue=config.get('WS_MESSAGE_QUEUE'),
    manage_session=config.get('WS_MANAGE_SESSION', False),
    engineio_logger=False,
    **options
)



# -*- coding: utf8 -*-
"""
.. module:: burpui.ext.sentry
    :platform: Unix
    :synopsis: Burp-UI external Sentry module.

.. moduleauthor:: Ziirish <hi+burpui@ziirish.me>

"""
from raven.contrib.flask import Sentry

sentry = Sentry()



# -*- coding: utf8 -*-
from six import iteritems

from .meta import meta_grants, BUIaclGrant
from .interface import BUIaclLoader
from ...utils import NOTIF_OK, NOTIF_WARN, NOTIF_ERROR, __


class ACLloader(BUIaclLoader):
    __doc__ = __("Uses the Burp-UI configuration file to load its rules.")
    section = name = 'BASIC:ACL'
    priority = 100

    def __init__(self, app=None):
        """See :func:`burpui.misc.acl.interface.BUIaclLoader.__init__`

        :param app: Application context
        :type app: :class:`burpui.server.BUIServer`
        """
        self.app = app
        self.conf = self.app.conf
        self.admins = [
            'admin'
        ]
        self.moderators = []
        self._groups = {}
        self._grants = {}
        self.first_setup = True
        self.moderator = {}
        self._acl = meta_grants
        self.conf_id = None
        self.meta_id = meta_grants.id
        meta_grants.register_backend(self.name, self)
        self.load_acl(True)

    def reload(self):
        self.load_acl(True)

    def load_acl(self, force=False):
        if not force and self.conf_id:
            if not self.conf.changed(self.conf_id):
                return False

        # our config changed or we were forced to reload our rules.
        # if the meta_grants didn't change, we reset them
        # if they changed, it means something else triggered a reset
        if not meta_grants.changed(self.meta_id) and not self.first_setup:
            meta_grants.reset(self.name)

        self.first_setup = False

        self.admins = [
            'admin'
        ]
        self.moderators = []
        self._grants = {}
        self.groups_def = {}

        adms = []
        mods = []

        def is_empty(data):
            return not data or data == [None]

        if self.section in self.conf.options:
            self.priority = self.conf.safe_get(
                'priority',
                'integer',
                section=self.section,
                defaults=self.priority
            )
            adms = self.conf.safe_get(
                'admin',
                'force_list',
                section=self.section
            )
            mods = self.conf.safe_get(
                '+moderator',
                'force_list',
                section=self.section
            )
            default_moderator = None
            if self.conf.get('STANDALONE'):
                default_moderator = '{"agents": {"rw": "local"}}'
            self.moderator = self.conf.safe_get(
                '@moderator',
                'force_string',
                section=self.section,
                defaults=default_moderator
            ) or {}
            meta_grants.set_moderator_grants(self.moderator)
            for opt in self.conf.options.get(self.section).keys():
                if opt in ['admin', '+moderator', 'priority', '@moderator']:
                    continue
                record = self.conf.safe_get(
                    opt,
                    'force_string',
                    section=self.section
                )

                self.logger.debug('record: {} -> {}'.format(opt, record))

                def _record(key):
                    if gname not in self.groups_def:
                        self.groups_def[gname] = {}
                    self.groups_def[gname][key] = parsed

                    return parsed

                if opt[0] == '+':
                    short = opt.lstrip('+')
                    gname = '@{}'.format(short)
                    parsed = meta_grants.set_group(gname, record)
                    self._groups[short] = parsed
                    _record('members')
                elif opt[0] == '@':
                    short = opt.lstrip('@')
                    if short not in self._groups:
                        self._groups[short] = []
                    gname = opt
                    parsed = record
                    _record('grants')
                    meta_grants.set_grant(gname, parsed)
                else:
                    self._grants[opt] = BUIaclGrant(opt, record).grants
                    meta_grants.set_grant(opt, record)

        if not is_empty(adms):
            self.admins = adms
        if not is_empty(mods):
            self.moderators = mods

        meta_grants.set_admin(self.admins)
        meta_grants.set_moderator(self.moderators)

        self.logger.debug('admins: {}'.format(self.admins))
        self.logger.debug('moderators: {}'.format(self.moderators))
        self.logger.debug('moderator grants: {}'.format(self.moderator))
        self.logger.debug('groups: {}'.format(self.groups_def))

        self.conf_id = self.conf.id
        self.meta_id = meta_grants.id

        return True

    def _setup_acl(self):
        """Setup ACL management"""
        if not self.conf.lookup_section(self.section):
            self.conf._refresh()

    def add_grant(self, name, content):
        """Add a grant"""
        if name[0] in ['+', '@']:
            message = "'{}' is not a valid grant name".format(name)
            self.logger.error(message)
            return False, message, NOTIF_ERROR
        self._setup_acl()
        if name in self.conf.options[self.section]:
            message = "grant '{}' already exists".format(name)
            self.logger.warning(message)
            return False, message, NOTIF_WARN
        self.conf.options[self.section][name] = content
        self.conf.options.write()
        self.load_acl(True)
        message = "grant '{}' successfully added".format(name)
        return True, message, NOTIF_OK

    def del_grant(self, name):
        """Delete a grant"""
        if name[0] in ['+', '@']:
            message = "'{}' is not a valid grant name".format(name)
            self.logger.error(message)
            return False, message, NOTIF_ERROR
        self._setup_acl()
        self.load_acl(True)
        if name not in self.conf.options[self.section]:
            message = "grant '{}' does not exist".format(name)
            self.logger.error(message)
            return False, message, NOTIF_ERROR
        del self.conf.options[self.section][name]
        self.conf.options.write()
        self.load_acl(True)
        if name in self.admins:
            self.del_admin(name)
        if name in self.moderators:
            self.del_moderator(name)
        for group, members in iteritems(self._groups):
            if name in members:
                self.del_group_member(group, name)
        message = "grant '{}' successfully removed".format(name)
        return True, message, NOTIF_OK

    def mod_grant(self, name, content):
        """Update a grant"""
        if name[0] in ['+', '@']:
            message = "'{}' is not a valid grant name".format(name)
            self.logger.error(message)
            return False, message, NOTIF_ERROR
        self._setup_acl()
        self.load_acl(True)
        if name not in self.conf.options[self.section]:
            message = "grant '{}' does not exist".format(name)
            self.logger.error(message)
            return False, message, NOTIF_WARN
        self.conf.options[self.section][name] = content
        self.conf.options.write()
        self.load_acl(True)
        message = "grant '{}' successfully modified".format(name)
        return True, message, NOTIF_OK

    def add_group(self, name, content):
        """Create a group"""
        self._setup_acl()
        name = '@{}'.format(name)
        if name == 'moderator':
            message = "'moderator' is a reserved name"
            self.logger.error(message)
            return False, message, NOTIF_ERROR
        if name in self.conf.options[self.section]:
            message = "group '{}' already exists".format(name)
            self.logger.warning(message)
            return False, message, NOTIF_WARN
        self.conf.options[self.section][name] = content
        self.conf.options.write()
        self.load_acl(True)
        message = "group '{}' successfully added".format(name)
        return True, message, NOTIF_OK

    def del_group(self, name):
        """Delete a group"""
        self._setup_acl()
        self.load_acl(True)
        gname = '@{}'.format(name)
        if name == 'moderator':
            message = "'moderator' is a reserved name"
            self.logger.error(message)
            return False, message, NOTIF_ERROR
        if gname not in self.conf.options[self.section]:
            message = "group '{}' does not exist".format(name)
            self.logger.error(message)
            return False, message, NOTIF_ERROR
        del self.conf.options[self.section][gname]
        gmembers = '+{}'.format(name)
        if gmembers in self.conf.options[self.section]:
            del self.conf.options[self.section][gmembers]
        self.conf.options.write()
        self.load_acl(True)
        if gname in self.admins:
            self.del_admin(gname)
        if gname in self.moderators:
            self.del_moderator(gname)
        for group, members in iteritems(self._groups):
            if gname in members:
                self.del_group_member(group, gname)
        message = "grant '{}' successfully removed".format(name)
        return True, message, NOTIF_OK

    def mod_group(self, name, content):
        """Update a group"""
        self._setup_acl()
        name = '@{}'.format(name)
        if name == 'moderator':
            message = "'moderator' is a reserved name"
            self.logger.error(message)
            return False, message, NOTIF_ERROR
        if name not in self.conf.options[self.section]:
            message = "group '{}' does not exist".format(name)
            self.logger.warning(message)
            return False, message, NOTIF_WARN
        self.conf.options[self.section][name] = content
        self.conf.options.write()
        self.load_acl(True)
        message = "group '{}' successfully modified".format(name)
        return True, message, NOTIF_OK

    def add_group_member(self, group, member):
        """Add a user to a group"""
        if group not in self._groups:
            message = "group '{}' does not exist".format(group)
            self.logger.warning(message)
            return False, message, NOTIF_WARN
        if member in self._groups[group]:
            message = "'{}' already in group '{}'".format(member, group)
            self.logger.warning(message)
            return False, message, NOTIF_WARN
        self._setup_acl()
        self._groups[group].append(member)
        gmembers = '+{}'.format(group)
        self.conf.options[self.section][gmembers] = self._groups[group]
        self.conf.options.write()
        self.load_acl(True)
        message = "'{}' added to group '{}'".format(member, group)
        return True, message, NOTIF_OK

    def del_group_member(self, group, member):
        """Remove a user from a group"""
        if group not in self._groups:
            message = "group '{}' does not exist".format(group)
            self.logger.warning(message)
            return False, message, NOTIF_WARN
        if member not in self._groups[group]:
            message = "'{}' not in group '{}'".format(member, group)
            self.logger.warning(message)
            return False, message, NOTIF_WARN
        self._setup_acl()
        self._groups[group].remove(member)
        gmembers = '+{}'.format(group)
        self.conf.options[self.section][gmembers] = self._groups[group] or ''
        self.conf.options.write()
        self.load_acl(True)
        message = "'{}' removed from group '{}'".format(member, group)
        return True, message, NOTIF_OK

    def add_moderator(self, member):
        """Add a moderator"""
        if member in self.moderators:
            message = "'{}' is already a moderator".format(member)
            self.logger.warning(message)
            return False, message, NOTIF_WARN
        self._setup_acl()
        self.moderators.append(member)
        self.conf.options[self.section]['+moderator'] = self.moderators
        self.conf.options.write()
        self.load_acl(True)
        message = "'{}' successfully added as moderator".format(member)
        return True, message, NOTIF_OK

    def del_moderator(self, member):
        """Delete a moderator"""
        if member not in self.moderators:
            message = "'{}' is not a moderator".format(member)
            self.logger.warning(message)
            return False, message, NOTIF_WARN
        self._setup_acl()
        self.moderators.remove(member)
        self.conf.options[self.section]['+moderator'] = self.moderators or ''
        self.conf.options.write()
        self.load_acl(True)
        message = "'{}' successfully removed from moderators".format(member)
        return True, message, NOTIF_OK

    def mod_moderator(self, grants):
        """Update moderator grants"""
        self._setup_acl()
        self.moderator = grants
        self.conf.options[self.section]['@moderator'] = grants
        self.conf.options.write()
        self.load_acl(True)
        message = "moderator grants updated"
        return True, message, NOTIF_OK

    def add_admin(self, member):
        """Add an admin"""
        if member in self.admins:
            message = "'{}' is already an admin".format(member)
            self.logger.warning(message)
            return False, message, NOTIF_WARN
        self._setup_acl()
        self.admins.append(member)
        self.conf.options[self.section]['admin'] = self.admins
        self.conf.options.write()
        self.load_acl(True)
        message = "'{}' successfully added as admin".format(member)
        return True, message, NOTIF_OK

    def del_admin(self, member):
        """Delete an admin"""
        if member not in self.admins:
            message = "'{}' is not an admin".format(member)
            self.logger.warning(message)
            return False, message, NOTIF_WARN
        self._setup_acl()
        self.admins.remove(member)
        self.conf.options[self.section]['admin'] = self.admins or ''
        self.conf.options.write()
        self.load_acl(True)
        message = "'{}' successfully removed from admins".format(member)
        return True, message, NOTIF_OK

    @property
    def acl(self):
        """Property to retrieve the backend"""
        if self._acl:
            self.load_acl()
            return self._acl
        return None  # pragma: no cover

    @property
    def grants(self):
        """Property to retrieve the list of grants"""
        return self._grants

    @property
    def groups(self):
        """Property to retrieve the list of groups with their members"""
        return self.groups_def



# -*- coding: utf8 -*-
"""
.. module:: burpui.misc.acl.meta
    :platform: Unix
    :synopsis: Burp-UI ACL meta definitions.

.. moduleauthor:: Ziirish <hi+burpui@ziirish.me>

"""
from .interface import BUIacl
from ...utils import make_list
from ...config import config
from ...ext.cache import cache

from six import iteritems, itervalues

import re
import json
import fnmatch


class BUImetaGrant(object):

    def _merge_data(self, d1, d2):
        """Merge data as list or dict recursively avoiding duplicates"""
        if not d1 and not d2:
            if isinstance(d1, dict) or isinstance(d2, dict):
                return {}
            return []
        if not d2:
            return d1
        if not d1:
            return d2
        if isinstance(d1, list) and isinstance(d2, list):
            return list(set(d1 + d2))
        if isinstance(d1, list) and not isinstance(d2, dict):
            if d2 in d1:
                return d1
            return d1 + make_list(d2)
        if isinstance(d2, list) and not isinstance(d1, dict):
            if d1 in d2:
                return d2
            return d2 + make_list(d1)
        if not isinstance(d1, dict) and not isinstance(d2, dict):
            if d1 == d2:
                return make_list(d1)
            else:
                return [d1, d2]

        res = d1
        for key2, val2 in iteritems(d2):
            if key2 in res:
                res[key2] = self._merge_data(val2, res[key2])
            else:
                res[key2] = val2
        return res

    def _parse_clients(self, data, mode=None, parent=None):
        agents = clients = []
        advanced = {}
        if isinstance(data, list):
            if mode:
                if parent and parent not in ['agents', 'clients', 'ro', 'rw']:
                    advanced[mode] = {parent: data}
                else:
                    advanced[mode] = {'clients': data}
            return data, agents, advanced
        if not isinstance(data, dict):
            if mode:
                if parent and parent not in ['agents', 'clients', 'ro', 'rw']:
                    advanced[mode] = {parent: make_list(data)}
                else:
                    advanced[mode] = {'clients': make_list(data)}
            return make_list(data), agents, advanced
        for key, val in iteritems(data):
            if key in ['agents', 'clients', 'ro', 'rw']:
                continue
            cl1, ag1, ad1 = self._parse_clients(val, parent=key)
            agents = self._merge_data(agents, ag1)
            clients = self._merge_data(clients, cl1)
            agents = self._merge_data(agents, key)
            advanced = self._merge_data(advanced, ad1)
            advanced = self._merge_data(advanced, {key: cl1})
            if mode:
                if parent and parent not in ['agents', 'clients', 'ro', 'rw']:
                    advanced = self._merge_data(advanced, {mode: {parent: cl1}})
                else:
                    advanced = self._merge_data(advanced, {mode: {key: cl1}})

        for key in ['clients', 'ro', 'rw']:
            md = None
            if key in data:
                if key in ['ro', 'rw']:
                    md = key
                cl2, ag2, ad2 = self._parse_clients(data[key], md, parent=key)
                agents = self._merge_data(agents, ag2)
                clients = self._merge_data(clients, cl2)
                if parent and parent not in ['agents', 'clients', 'ro', 'rw']:
                    ro = ad2.get('ro')
                    rw = ad2.get('rw')
                    if ro and 'clients' in ro:
                        ro[parent] = ro['clients']
                        del ro['clients']
                        ad2['ro'] = ro
                    if rw and 'clients' in rw:
                        rw[parent] = rw['clients']
                        del rw['clients']
                        ad2['rw'] = rw
                advanced = self._merge_data(advanced, ad2)

        if 'agents' in data:
            ag3, cl3, ad3 = self._parse_agents(data['agents'])
            agents = self._merge_data(agents, ag3)
            clients = self._merge_data(clients, cl3)
            advanced = self._merge_data(advanced, ad3)

        return make_list(clients), make_list(agents), advanced

    def _parse_agents(self, data, mode=None):
        agents = clients = []
        advanced = {}
        if isinstance(data, list):
            if mode:
                advanced[mode] = {'agents': data}
            return data, clients, advanced
        if not isinstance(data, dict):
            if mode:
                advanced[mode] = {'agents': make_list(data)}
            return make_list(data), clients, advanced
        for key, val in iteritems(data):
            if key in ['agents', 'clients', 'ro', 'rw']:
                continue
            cl1, ag1, ad1 = self._parse_clients(data)
            agents = self._merge_data(agents, ag1)
            clients = self._merge_data(clients, cl1)
            agents = self._merge_data(agents, key)
            advanced = self._merge_data(advanced, ad1)
            if mode:
                advanced = self._merge_data(advanced, {mode: ad1})
            # FIXME: why did I do that?
            # advanced = self._merge_data(advanced, {key: cl1})
            # if mode:
            #     advanced = self._merge_data(advanced, {mode: {key: cl1}})

        for key in ['agents', 'ro', 'rw']:
            md = None
            if key in data:
                if key in ['ro', 'rw']:
                    md = key
                ag2, cl2, ad2 = self._parse_agents(data[key], md)
                agents = self._merge_data(agents, ag2)
                clients = self._merge_data(clients, cl2)
                advanced = self._merge_data(advanced, ad2)

        if 'clients' in data:
            cl3, ag3, ad3 = self._parse_clients(data['clients'])
            agents = self._merge_data(agents, ag3)
            clients = self._merge_data(clients, cl3)
            advanced = self._merge_data(advanced, ad3)

        return make_list(agents), make_list(clients), advanced


class BUIgrantHandler(BUImetaGrant, BUIacl):
    """This class is here to handle grants in a generic way.
    It will automatically merge grants from various backends that register to it
    """
    _id = 1
    _gp_admin_name = '@BUIADMINRESERVED'
    _gp_moderator_name = '@moderator'
    _gp_hidden = set([str(_gp_admin_name[1:]), str(_gp_moderator_name[1:])])

    _grants = {}
    _groups = {}

    _options = {}
    _backends = {}

    _name = 'meta_grant'

    @property
    def id(self):
        """current handler id, used to detect configuration changes"""
        return self._id

    @property
    def grants(self):
        """grants managed by our handler"""
        return self._grants

    @property
    def groups(self):
        """groups managed by our handler"""
        return self._groups

    @property
    def options(self):
        """options of our ACL engine"""
        return self._options

    @options.setter
    def options(self, value):
        """set the options of our engine"""
        self._options = value
        if self._options.get('legacy'):
            self._options['extended'] = False

    def changed(self, sid):
        """detect a configuration change"""
        return sid != self._id

    def reset(self, reset_from):
        """a configuration change occurred, we reload our grants and groups"""
        self._grants.clear()
        self._groups.clear()
        self._reset_cached()
        self._id += 1
        for name, backend in iteritems(self._backends):
            if name == reset_from:
                continue
            backend.reload()

    def opt(self, key, default=False):
        """access a given option"""
        if key not in self.options:
            return default
        return self.options.get(key)

    def register_backend(self, name, backend):
        """register a new ACL backend

        :param name: Backend name
        :type name: str

        :param backend: ACL Backend
        :type backend: :class:`burpui.misc.acl.interface.BUIaclLoader`
        """
        self._backends[name] = backend

    def set_grant(self, name, grant):
        """parse and set the given grants"""
        if name in self._grants:
            return self._grants[name].add_grants(grant)
        self._grants[name] = BUIaclGrant(name, grant)
        return self._grants[name].grants

    def set_group(self, name, members):
        """parse and set the given group"""
        if name in self._groups:
            return self._groups[name].add_members(members)
        self._groups[name] = BUIaclGroup(name, members)
        return self._groups[name].members

    def set_admin(self, admins):
        """parse and set the admins"""
        self.set_group(self._gp_admin_name, admins)

    def set_moderator(self, moderators):
        """parse and set the moderators"""
        self.set_group(self._gp_moderator_name, moderators)

    def set_moderator_grants(self, grants):
        """parse and set the moderators grants"""
        self.set_grant(self._gp_moderator_name, grants)

    def get_member_groups(self, member):
        groups = []
        for group in itervalues(self._groups):
            (ret, inh) = group.is_member(member)
            if ret and group.name not in self._gp_hidden:
                groups.append((group.name, inh))
        return groups

    def _gen_key(self, username):
        return '{}-{}'.format(self._name, username)

    def _set_cached(self, username, value):
        key = self._gen_key(username)
        return cache.cache.set(key, value)

    def _get_cached(self, username):
        key = self._gen_key(username)
        return cache.cache.get(key)

    def _reset_cached(self):
        cache.clear()

    def _is_cached(self, username):
        key = self._gen_key(username)
        return cache.cache.has(key)

    def _extract_grants(self, username, parent=None):
        if not self._is_cached(username):
            data = {}

            if username in self.grants:
                grants = self.grants[username].grants
            else:
                grants = []

            clients, agents, advanced = self._parse_clients(grants)
            data['clients'] = clients
            data['agents'] = agents
            data['advanced'] = [advanced] if advanced else []

            def __merge_grants_with(grp, prt):
                data2 = self._extract_grants(grp, prt)
                data['clients'] = self._merge_data(
                    data['clients'],
                    data2['clients']
                )
                data['agents'] = self._merge_data(
                    data['agents'],
                    data2['agents']
                )
                tmp = data2['advanced']
                if tmp:
                    data['advanced'] += tmp

            # moderator is also a group
            for gname, group in iteritems(self.groups):
                # no grants need to be parsed for admins
                if gname == self._gp_admin_name:
                    continue
                (ret, _) = group.is_member(username)
                if not parent:
                    parent = set([username])
                elif isinstance(parent, set):
                    parent.add(username)
                if ret and gname != username and parent and gname not in parent:
                    __merge_grants_with(gname, parent)

            self._set_cached(username, data)
            return data
        return self._get_cached(username)

    def _extract_clients(self, username):
        ret = self._extract_grants(username)
        return ret.get('clients', [])

    def _extract_agents(self, username):
        ret = self._extract_grants(username)
        return ret.get('agents', [])

    def _extract_advanced(self, username, idx=None):
        ret = self._extract_grants(username).get('advanced', [])
        if idx is not None:
            return ret[idx]
        if self.opt('inverse_inheritance'):
            return reversed(ret)
        return ret

    def _extract_advanced_mode(self, username, mode, kind, idx):
        return self._extract_advanced(username, idx).get(mode, {}).get(kind, [])

    def _client_match(self, username, client):
        clients = self._extract_clients(username)
        if not clients:
            return None

        if self.opt('extended'):
            matches = []
            for exp in clients:
                regex = fnmatch.translate(exp)
                if re.match(regex, client):
                    matches.append(exp)
            return matches if matches else False
        else:
            return [client] if client in clients else False

    def _server_match(self, username, server):
        servers = self._extract_agents(username)
        if not servers:
            return None

        if self.opt('extended'):
            matches = []
            for exp in servers:
                regex = fnmatch.translate(exp)
                if re.match(regex, server):
                    matches.append(exp)
            return matches if matches else False
        else:
            return [server] if server in servers else False

    # implement BUIacl methods

    def is_admin(self, username):
        """See :func:`burpui.misc.acl.interface.BUIacl.is_admin`"""
        if self._gp_admin_name in self._groups:
            return self._groups[self._gp_admin_name].is_member(username)
        return False, None

    def is_moderator(self, username):
        """See :func:`burpui.misc.acl.interface.BUIacl.is_moderator`"""
        if self._gp_moderator_name in self._groups:
            return self._groups[self._gp_moderator_name].is_member(username)
        return False, None

    def is_client_rw(self, username=None, client=None, server=None):
        """See :func:`burpui.misc.acl.interface.BUIacl.is_client_rw`"""
        if not username or not client:  # pragma: no cover
            return False

        (is_admin, _) = self.is_admin(username)

        ret = is_admin or self.opt('assume_rw', True) or self.opt('legacy')

        if self.is_client_allowed(username, client, server):
            # legacy mode: assume rw for everyone
            if self.opt('legacy'):
                return True
            client_match = self._client_match(username, client)
            advanced = self._extract_advanced(username)

            if client_match is None and username == client:
                client_match = [username]

            if server:
                server_match = self._server_match(username, server)

                if not server_match and not client_match:
                    return is_admin or self.opt('assume_rw', True)

                for adv in advanced:
                    for adv2 in advanced:
                        # the whole agent is rw and we did not find explicit entry for
                        # client_match
                        if client_match is False:
                            if server_match and \
                                    any([x in adv.get('rw', {}) or
                                         x in adv.get('rw', {}).get('agents', [])
                                         for x in server_match]):
                                return True
                            if server in adv.get('rw', {}) or \
                                    server in adv.get('rw', {}).get('agents', []):
                                return True

                        if server_match and \
                                any([x in adv.get('rw', {}) or
                                     x in adv.get('rw', {}).get('agents', [])
                                     for x in server_match]):
                            # both agent and client are defined as rw
                            if client_match and \
                                any([x in adv2.get('rw', []) or
                                     x in adv2.get('rw', {}).get('clients', []) or
                                     any([x in adv2.get('rw', {}).get(y, [])
                                          for y in server_match])
                                     for x in client_match
                                     ]):
                                return True

                            # the agent is rw but the client is explicitly defined as ro
                            if client_match and \
                                any([x in adv2.get('ro', []) or
                                     x in adv2.get('ro', {}).get('clients', []) or
                                     any([x in adv2.get('ro', {}).get(y, [])
                                          for y in server_match])
                                     for x in client_match
                                     ]):
                                return False

                        if server_match and \
                                any([x in adv.get('ro', {}) or
                                     x in adv.get('ro', {}).get('agents', [])
                                     for x in server_match]):
                            # the agent is ro, but the client is explicitly defined as rw
                            if client_match and \
                                any([x in adv2.get('rw', {}).get('clients', []) or
                                     x in adv2.get('rw', []) or
                                     any([x in adv2.get('rw', {}).get(y, [])
                                          for y in server_match])
                                     for x in client_match
                                     ]):
                                return True

                            # both server and client are explicitly defined as ro
                            if client_match and \
                                any([x in adv2.get('ro', {}).get('clients', []) or
                                     x in adv2.get('ro', []) or
                                     any([x in adv2.get('ro', {}).get(y, [])
                                         for y in server_match])
                                     for x in client_match
                                     ]):
                                return False

            for adv in advanced:
                # client is explicitly defined as rw
                rw_clients = adv.get('rw', {}).get('clients', [])
                if client_match and \
                        any([x in rw_clients for x in client_match]):
                    return True

                if client and \
                        client in rw_clients:
                    return True

                # client is explicitly defined as ro
                ro_clients = adv.get('ro', {}).get('clients', [])
                if client_match and \
                        any([x in ro_clients for x in client_match]):
                    return False

                if client and \
                        client in ro_clients:
                    return False

        return ret

    def is_client_allowed(self, username=None, client=None, server=None):
        """See :func:`burpui.misc.acl.interface.BUIacl.is_client_allowed`"""
        if not username or not client:  # pragma: no cover
            return False

        (is_admin, _) = self.is_admin(username)
        client_match = self._client_match(username, client)

        if client_match is None and username == client:
            client_match = [username]
        elif not client_match:
            client_match = False

        if server:
            server_match = self._server_match(username, server)
            if server_match is not None or self.opt('legacy'):
                if not server_match:
                    return is_admin

                advanced = self._extract_advanced(username)
                if self.opt('implicit_link', True) and not advanced:
                    advanced = False

                if advanced is not False:
                    for idx, adv in enumerate(advanced):
                        if all([x not in adv for x in server_match]) and \
                                any([x in y
                                     for x in server_match
                                     for y in self._extract_advanced_mode(username, 'ro', 'agents', idx)
                                     ]) or \
                                any([x in y
                                     for x in server_match
                                     for y in self._extract_advanced_mode(username, 'rw', 'agents', idx)
                                     ]):
                            return True

                        tmp = set(adv.get(server, []))
                        for srv in server_match:
                            tmp |= set(adv.get(srv, []))
                        adv2 = list(tmp)
                        if client_match is not False and \
                                (any([x in adv2 for x in client_match]) or
                                 client in adv2):
                            return True

                    return False

        return client_match is not False or is_admin

    def is_server_rw(self, username=None, server=None):
        """See :func:`burpui.misc.acl.interface.BUIacl.is_server_rw`"""
        if not username:  # pragma: no cover
            return False

        # special case single-agent mode
        if not server and config.get('STANDALONE'):
            server = 'local'

        (is_admin, _) = self.is_admin(username)

        ret = is_admin or self.opt('assume_rw', True) or self.opt('legacy')

        if self.is_server_allowed(username, server):
            server_match = self._server_match(username, server)
            if not server_match:
                return is_admin or self.opt('assume_rw', True)

            advanced = self._extract_advanced(username)

            for adv in advanced:
                # server is explicitly defined as ro
                if any([x in adv.get('ro', {}).get('agents', []) for x in server_match]):
                    return False

                # server is explicitly defined as rw
                if any([x in adv.get('rw', {}).get('agents', []) for x in server_match]):
                    return True

        return ret

    def is_server_allowed(self, username=None, server=None):
        """See :func:`burpui.misc.acl.interface.BUIacl.is_server_allowed`"""
        if not username or not server:
            return False

        server_match = self._server_match(username, server)
        (is_admin, _) = self.is_admin(username)

        if server_match is None and self.opt('legacy'):
            server_match = False

        return server_match is not False or is_admin


class BUIaclGroup(object):
    """The :class:`burpui.misc.acl.interface.BUIaclGroup` class is used to
    represent a Group"""

    def __init__(self, name, members=None):
        self._name = name
        self._set_members(members)
        self.has_subgroups = -1

    def _parse_members(self, members):
        # we support only lists
        if members and ',' in members and not isinstance(members, list):
            parsed = [x.strip() for x in members.split(',')]
        else:
            parsed = make_list(members)
        return parsed

    def _set_members(self, members):
        self._members = set(self._parse_members(members))

    def add_members(self, new_members):
        new_members = self._parse_members(new_members)
        self._members = self._members | set(new_members)
        # reset the flag
        self.has_subgroups = -1
        return new_members

    def del_members(self, members_remove):
        members_remove = self._parse_members(members_remove)
        self._members = self._members - set(members_remove)
        # reset the flag
        self.has_subgroups = -1

    def is_member(self, member, parent=None):
        inherit = set()
        ret = member in self._members
        if not ret and (self.has_subgroups > 0 or self.has_subgroups == -1):
            self.has_subgroups = 0
            for mem in self.members:
                # avoid infinite loop with mutual inheritance
                if parent and mem in parent:
                    continue
                if mem.startswith('@'):
                    self.has_subgroups += 1
                    if mem in meta_grants._groups:
                        if parent:
                            parent.append(mem)
                        else:
                            parent = [mem]
                        (ret, inh2) = meta_grants._groups[mem].is_member(member, parent=parent)
                        if ret:
                            for subinh in inh2:
                                inherit.add(subinh)
                            inherit.add(mem)
                            # no break, we may have other inheritance at the level
        return ret, list(inherit)

    @property
    def name(self):
        if self._name and any(self._name.startswith(x) for x in ['@', '+']):
            return str(self._name[1:])
        return self._name

    @property
    def members(self):
        return list(self._members)


class BUIaclGrant(BUImetaGrant):
    """The :class:`burpui.misc.acl.interface.BUIaclGrant` class is used to
    represent a Grant"""

    def __init__(self, name, grants):
        self._name = name
        self._grants = self._parse_grants(grants)

    def _parse_grants(self, grants):
        try:
            ret = json.loads(grants)
        except (ValueError, TypeError):
            # handle empty/missing grants
            if not grants:
                return []
            # ignore mal-formatted json
            if any([x in grants for x in ['{', '}', '[', ']']]):
                ret = None
            elif grants and ',' in grants:
                ret = [x.rstrip() for x in grants.split(',')]
            else:
                ret = make_list(grants)
        return ret

    @property
    def name(self):
        if self._name and any(self._name.startswith(x) for x in ['@', '+']):
            return str(self._name[1:])
        return self._name

    @property
    def grants(self):
        return self._grants

    @property
    def grants_raw(self):
        return json.dumps(self._grants)

    def add_grants(self, grants):
        parsed = self._parse_grants(grants)
        self._grants = self._merge_data(self._grants, parsed)
        return parsed


meta_grants = BUIgrantHandler()



"""user prefs

Revision ID: 56de018f4d88
Revises: fc07e3fa0086
Create Date: 2017-02-12 14:51:38.147422

"""

# revision identifiers, used by Alembic.
revision = '56de018f4d88'
down_revision = 'fc07e3fa0086'

from alembic import op
import sqlalchemy as sa


def upgrade():
    ### commands auto generated by Alembic - please adjust! ###
    op.create_table('pref',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('user', sa.String(length=256), nullable=True),
    sa.Column('key', sa.String(length=256), nullable=False),
    sa.Column('value', sa.String(length=256), nullable=True),
    sa.PrimaryKeyConstraint('id')
    )
    ### end Alembic commands ###


def downgrade():
    ### commands auto generated by Alembic - please adjust! ###
    op.drop_table('pref')
    ### end Alembic commands ###



# -*- coding: utf8 -*-
"""

.. module:: burpui.extensions
    :platform: Unix
    :synopsis: Burp-UI extensions module.

.. moduleauthor:: Ziirish <hi+burpui@ziirish.me>
"""
import os
import re
import warnings

from ._compat import PY3, to_unicode

if PY3:  # pragma: no cover
    basestring = str


def parse_db_setting(string):
    parts = re.search(
        '(?:(?P<backend>\w+)(?:\+(?P<driver>\w+))?://)?'
        '(?:(?P<user>\w+)(?::?(?P<pass>.+))?@)?'
        '(?P<host>[\w_.-]+):?(?P<port>\d+)?(?:/(?P<db>\w+))?',
        string
    )
    if not parts:  # pragma: no cover
        raise ValueError('Unable to parse the db: "{}"'.format(string))
    back = parts.group('backend') or ''
    user = parts.group('user') or None
    pwd = parts.group('pass') or None
    host = parts.group('host') or ''
    port = parts.group('port') or ''
    db = parts.group('db') or ''
    return (back, user, pwd, host, port, db)


def get_redis_server(myapp):
    host = 'localhost'
    port = 6379
    pwd = None
    if myapp.redis and myapp.redis.lower() != 'none':
        try:
            back, user, pwd, host, port, db = parse_db_setting(myapp.redis)
            host = host or 'localhost'
            try:
                port = int(port)
            except (ValueError, IndexError):
                port = 6379
        except ValueError:  # pragma: no cover
            pass
    return host, port, pwd


def create_db(myapp, cli=False, unittest=False, create=True, celery_worker=False):
    """Create the SQLAlchemy instance if possible

    :param myapp: Application context
    :type myapp: :class:`burpui.server.BUIServer`
    """
    if myapp.config['WITH_SQL']:
        try:
            from .ext.sql import db
            from sqlalchemy.exc import OperationalError
            from sqlalchemy_utils.functions import database_exists
            myapp.config['SQLALCHEMY_TRACK_MODIFICATIONS'] = False
            if not database_exists(myapp.config['SQLALCHEMY_DATABASE_URI']) and \
                    not cli and not unittest and not celery_worker:
                if create:  # pragma: no cover
                    import subprocess
                    local = os.path.join(os.getcwd(), '..', 'tools', 'bui-manage')
                    buimanage = local if os.path.exists(local) else 'bui-manage'
                    cmd = [
                        buimanage,
                        '-c',
                        myapp.config['CFG'],
                        '-l',
                        os.devnull,
                        'db',
                        'upgrade'
                    ]
                    upgd = subprocess.Popen(
                        cmd,
                        stdout=subprocess.PIPE,
                        stderr=subprocess.STDOUT
                    )
                    (out, _) = upgd.communicate()
                    if upgd.returncode != 0:
                        myapp.logger.error(
                            'Disabling SQL support because '
                            'something went wrong while setting up the '
                            'database:\n{}'.format(out)
                        )
                        myapp.config['WITH_SQL'] = False
                        return None
                    return create_db(myapp, cli, unittest, False)
                else:  # pragma: no cover
                    myapp.logger.error(
                        'Database not found, disabling SQL support'
                    )
                    myapp.config['WITH_SQL'] = False
                    return None

            back = parse_db_setting(myapp.config['SQLALCHEMY_DATABASE_URI'])[0]

            if 'mysql' in back:  # pragma: no cover
                # optimize SQL pools for MySQL driver
                myapp.config['SQLALCHEMY_POOL_SIZE'] = 20
                myapp.config['SQLALCHEMY_POOL_RECYCLE'] = 600

            db.init_app(myapp)
            if not cli and not unittest and not celery_worker:  # pragma: no cover
                with myapp.app_context():
                    try:
                        import subprocess

                        # get the current revision from alembic_version
                        res = db.engine.execute(
                            'select version_num from alembic_version'
                        )
                        if not res:
                            raise Exception(
                                'Alembic does not seem to be setup'
                            )
                        current = None
                        for row in res:
                            current = to_unicode(row['version_num'])
                            break

                        # get current head using alembic/FLask-Migrate
                        local = os.path.join(
                            os.getcwd(),
                            'tools',
                            'bui-manage'
                        )
                        buimanage = local if os.path.exists(local) \
                            else 'bui-manage'
                        cmd = [
                            buimanage,
                            '-c',
                            myapp.config['CFG'],
                            '-l',
                            os.devnull,
                            'db',
                            'heads'
                        ]
                        rev = subprocess.Popen(
                            cmd,
                            stdout=subprocess.PIPE,
                            stderr=subprocess.STDOUT
                        )
                        (out, _) = rev.communicate()
                        if rev.returncode != 0:
                            raise Exception(
                                'something went wrong while setting up the '
                                'database:\n{}'.format(out)
                            )

                        latest = to_unicode(out).split()[0]

                        # now we compare the revision numbers
                        if latest != current:
                            myapp.logger.critical(
                                'Your database seems out of sync ({} != {}), '
                                'you may want to run \'bui-manage db '
                                'upgrade\'.'.format(latest, current)
                            )
                            myapp.logger.critical(
                                'Disabling SQL support for now.'
                            )
                            myapp.config['WITH_SQL'] = False
                            return None

                    except (OperationalError, Exception) as exp:
                        err = str(exp)
                        if 'no such table' in err:
                            myapp.logger.critical(
                                'Your database seems out of sync, you may want '
                                'to run \'bui-manage db upgrade\'.'
                            )
                        else:
                            myapp.logger.critical(
                                'Something seems to be wrong with your setup: '
                                '{}'.format(err)
                            )

                        myapp.logger.critical('Disabling SQL support for now.')
                        myapp.config['WITH_SQL'] = False
                        return None

            # If we are here, it means everything is alright
            return db

        except ImportError:  # pragma: no cover
            myapp.logger.critical(
                'Unable to load requirements, you may want to run \'pip '
                'install "burp-ui[sql]"\'.\nDisabling SQL support for now.'
            )
            myapp.config['WITH_SQL'] = False
        except OperationalError as exp:  # pragma: no cover
            myapp.logger.critical(
                'unable to contact database: {}\nDisabling SQL '
                'support.'.format(exp)
            )
            myapp.config['WITH_SQL'] = False

    return None


def create_websocket(myapp, websocket_server=False, celery_worker=False,
                     gunicorn=False, cli=False):
    """Create the websocket server if possible

    :param myapp: Application context
    :type myapp: :class:`burpui.server.BUIServer`
    """
    if cli and not websocket_server:
        return
    broker = myapp.ws_broker
    if broker is not False:
        if not broker or broker is True:
            broker = 'redis'
    if broker and broker.lower() != 'none':
        host, oport, pwd = get_redis_server(myapp)
        odb = 4
        if broker.lower() not in ['default', 'redis']:
            try:
                (_, _, pwd, host, port, db) = parse_db_setting(myapp.use_celery)
                if not port:
                    port = oport
                if not db:
                    db = odb
                else:
                    try:
                        db = int(db)
                    except ValueError:
                        db = odb
            except ValueError:
                pass
        else:
            port = oport
            db = odb
        if pwd:
            redis_url = 'redis://:{}@{}:{}/{}'.format(pwd, host, port, db)
        else:
            redis_url = 'redis://{}:{}/{}'.format(host, port, db)
        myapp.config['WS_MESSAGE_QUEUE'] = redis_url
    myapp.config['WS_MANAGE_SESSION'] = not myapp.config.get('WITH_SRV_SESSION', False)
    if os.getenv('BUI_MODE') == 'celery':
        myapp.config['WS_ASYNC_MODE'] = 'threading'
    # myapp.config['WS_ASYNC_MODE'] = 'threading' if not gunicorn else None

    if celery_worker:
        return

    # if you are not a celery worker, we can patch the flask server
    try:
        from .ext.ws import socketio
        socketio.init_app(
            myapp,
            message_queue=myapp.config.get('WS_MESSAGE_QUEUE'),
            manage_session=myapp.config.get('WS_MANAGE_SESSION', False)
        )
        myapp.config['WS_AVAILABLE'] = True
    except ImportError:
        myapp.config['WS_AVAILABLE'] = False

    # Now load the namespaces
    if myapp.config['WITH_WS'] or websocket_server:
        from .ws.namespace import BUINamespace
        socketio.on_namespace(BUINamespace('/ws'))


def create_celery(myapp, warn=True):
    """Create the Celery app if possible

    :param myapp: Application context
    :type myapp: :class:`burpui.server.BUIServer`
    """
    if myapp.config['WITH_CELERY']:  # pragma: no cover
        from .exceptions import BUIserverException
        host, oport, pwd = get_redis_server(myapp)
        odb = 2
        if isinstance(myapp.use_celery, basestring):
            try:
                (_, _, pwd, host, port, db) = parse_db_setting(myapp.use_celery)
                if not port:
                    port = oport
                if not db:
                    db = odb
                else:
                    try:
                        db = int(db)
                    except ValueError:
                        db = odb
            except ValueError:
                pass
        else:
            db = odb
            port = oport
        if pwd:
            redis_url = 'redis://:{}@{}:{}/{}'.format(pwd, host, port, db)
        else:
            redis_url = 'redis://{}:{}/{}'.format(host, port, db)
        myapp.config['CELERY_BROKER_URL'] = myapp.config['BROKER_URL'] = \
            redis_url
        myapp.config['CELERY_RESULT_BACKEND'] = redis_url

        from .ext.async import celery
        celery.conf.update(myapp.config)

        if not hasattr(celery, 'flask_app'):
            celery.flask_app = myapp

        TaskBase = celery.Task

        class ContextTask(TaskBase):
            abstract = True

            def __call__(self, *args, **kwargs):
                with myapp.app_context():
                    try:
                        return TaskBase.__call__(self, *args, **kwargs)
                    except BUIserverException:
                        # ignore unhandled exceptions in the celery worker
                        pass

        celery.Task = ContextTask

        # may fail in case redis is not running (this can happen while running
        # the bui-manage script)
        try:
            from .tasks import force_scheduling_now
            force_scheduling_now()
        except:  # pragma: no cover
            pass

        return celery

    if warn:  # pragma: no cover
        message = 'Something went wrong while initializing celery worker.\n' \
                  'Maybe it is not enabled in your conf ' \
                  '({}).'.format(myapp.config['CFG'])
        warnings.warn(
            message,
            RuntimeWarning
        )

    return None



# -*- coding: utf8 -*-
"""
.. module:: burpui.api.clients
    :platform: Unix
    :synopsis: Burp-UI clients api module.

.. moduleauthor:: Ziirish <hi+burpui@ziirish.me>

"""
from . import api, cache_key, force_refresh
from ..server import BUIServer  # noqa
from .custom import fields, Resource
from .client import ClientLabels
from ..ext.cache import cache
from ..exceptions import BUIserverException
from ..decorators import browser_cache

from six import iteritems
from flask import current_app
from flask_login import current_user

bui = current_app  # type: BUIServer
ns = api.namespace('clients', 'Clients methods')


@ns.route('/running',
          '/<server>/running',
          '/running/<client>',
          '/<server>/running/<client>',
          endpoint='running_clients')
@ns.doc(
    params={
        'server': 'Which server to collect data from when in multi-agent mode',
        'client': 'Client name',
    },
)
class RunningClients(Resource):
    """The :class:`burpui.api.clients.RunningClients` resource allows you to
    retrieve a list of clients that are currently running a backup.

    This resource is part of the :mod:`burpui.api.clients` module.

    An optional ``GET`` parameter called ``serverName`` is supported when running
    in multi-agent mode.
    """
    parser = ns.parser()
    parser.add_argument('serverName', help='Which server to collect data from when in multi-agent mode')

    @ns.expect(parser)
    def get(self, client=None, server=None):
        """Returns a list of clients currently running a backup

        **GET** method provided by the webservice.

        The *JSON* returned is:
        ::

            [ 'client1', 'client2' ]


        The output is filtered by the :mod:`burpui.misc.acl` module so that you
        only see stats about the clients you are authorized to see.

        :param server: Which server to collect data from when in multi-agent mode
        :type server: str

        :param client: Ask a specific client in order to know if it is running a backup
        :type client: str

        :returns: The *JSON* described above.
        """
        server = server or self.parser.parse_args()['serverName']
        return self._running_clients(None, client, server)

    def _running_clients(self, res, client, server):
        if client:
            if not current_user.is_anonymous and \
                    not current_user.acl.is_admin() and \
                    not current_user.acl.is_client_allowed(client, server):
                return []

            if bui.client.is_backup_running(client, server):
                return [client]
            else:
                return []

        running = res or bui.client.is_one_backup_running(server)
        # Manage ACL
        if not current_user.is_anonymous and not current_user.acl.is_admin():
            if isinstance(running, dict):
                ret = {}

                def __extract_running_clients(serv):
                    try:
                        clients = [x['name'] for x in bui.client.get_all_clients(serv)]
                    except BUIserverException:
                        clients = []
                    allowed = [x for x in clients if current_user.acl.is_client_allowed(x, serv)]
                    return [x for x in running[serv] if x in allowed]

                if server:
                    return __extract_running_clients(server)

                for serv in bui.client.servers:
                    ret[serv] = __extract_running_clients(serv)
                return ret
            else:
                try:
                    clients = [x['name'] for x in bui.client.get_all_clients(server)]
                except BUIserverException:
                    clients = []
                allowed = [x for x in clients if current_user.acl.is_client_allowed(x, server)]
                running = [x for x in running if x in allowed]
        elif server and isinstance(running, dict):
            return running.get(server, [])
        return running


@ns.route('/backup-running',
          '/<server>/backup-running',
          endpoint='running_backup')
@ns.doc(
    params={
        'server': 'Which server to collect data from when in multi-agent mode',
    }
)
class RunningBackup(Resource):
    """The :class:`burpui.api.clients.RunningBackup` resource allows you to
    access the status of the server in order to know if there is a running
    backup currently.

    This resource is part of the :mod:`burpui.api.clients` module.
    """
    running_fields = ns.model('Running', {
        'running': fields.Boolean(required=True, description='Is there a backup running right now'),
    })

    @ns.marshal_with(running_fields, code=200, description='Success')
    def get(self, server=None):
        """Tells if a backup is running right now

        **GET** method provided by the webservice.

        The *JSON* returned is:
        ::

            {
                "running": false
            }


        The output is filtered by the :mod:`burpui.misc.acl` module so that you
        only see stats about the clients you are authorized to.

        :param server: Which server to collect data from when in multi-agent mode
        :type server: str

        :returns: The *JSON* described above.
        """
        return {
            'running': self._is_one_backup_running(
                bui.client.is_one_backup_running(server),
                server
            )
        }

    def _is_one_backup_running(self, res, server):
        """Check if a backup is running"""
        # Manage ACL
        if not current_user.is_anonymous and not current_user.acl.is_admin():
            if isinstance(res, dict):
                new = {}
                for serv in bui.client.servers:
                    try:
                        clients = [x['name'] for x in bui.client.get_all_clients(serv)]
                    except BUIserverException:
                        clients = []
                    allowed = [x for x in clients if current_user.acl.is_client_allowed(x, serv)]
                    new[serv] = [x for x in res[serv] if x in allowed]
                res = new
            else:
                try:
                    clients = [x['name'] for x in bui.client.get_all_clients(server)]
                except BUIserverException:
                    clients = []
                allowed = [x for x in clients if current_user.acl.is_client_allowed(x, server)]
                res = [x for x in res if x in allowed]
        running = False
        if isinstance(res, dict):
            for (_, run) in iteritems(res):
                running = running or (len(run) > 0)
                if running:
                    break
        else:
            running = len(res) > 0

        return running


@ns.route('/report',
          '/<server>/report',
          endpoint='clients_report')
@ns.doc(
    params={
        'server': 'Which server to collect data from when in multi-agent mode',
    },
)
class ClientsReport(Resource):
    """The :class:`burpui.api.clients.ClientsReport` resource allows you to
    access general reports about your clients.

    This resource is part of the :mod:`burpui.api.clients` module.

    An optional ``GET`` parameter called ``serverName`` is supported when running
    in multi-agent mode.
    """
    parser = ns.parser()
    parser.add_argument('serverName', help='Which server to collect data from when in multi-agent mode')
    parser.add_argument('limit', type=int, default=8, help='Number of elements to return')
    parser.add_argument('aggregation', help='What aggregation to operate', default='number', choices=('number', 'files', 'size', 'none'))

    translation = {
        'number': 'number',
        'files': 'total',
        'size': 'totsize',
    }

    stats_fields = ns.model('ClientsStats', {
        'total': fields.Integer(required=True, description='Number of files', default=0),
        'totsize': fields.Integer(required=True, description='Total size occupied by all the backups of this client', default=0),
        'os': fields.String(required=True, description='OS of the client', default='unknown'),
    })
    client_fields = ns.model('ClientsReport', {
        'name': fields.String(required=True, description='Client name'),
        'stats': fields.Nested(stats_fields, required=True),
    })
    backup_fields = ns.model('ClientsBackup', {
        'name': fields.String(required=True, description='Client name'),
        'number': fields.Integer(required=True, description='Number of backups on this client', default=0),
    })
    report_fields = ns.model('Report', {
        'backups': fields.Nested(backup_fields, as_list=True, required=True),
        'clients': fields.Nested(client_fields, as_list=True, required=True),
    })

    @cache.cached(timeout=1800, key_prefix=cache_key, unless=force_refresh)
    @ns.marshal_with(report_fields, code=200, description='Success')
    @ns.expect(parser)
    @ns.doc(
        responses={
            403: 'Insufficient permissions',
            500: 'Internal failure',
        },
    )
    @browser_cache(1800)
    def get(self, server=None):
        """Returns a global report about all the clients of a given server

        **GET** method provided by the webservice.

        The *JSON* returned is:
        ::

            {
              "backups": [
                {
                  "name": "client1",
                  "number": 15
                },
                {
                  "name": "client2",
                  "number": 1
                }
              ],
              "clients": [
                {
                  "name": "client1",
                  "stats": {
                    "total": 296377,
                    "totsize": 57055793698,
                    "os": "unknown"
                  }
                },
                {
                  "name": "client2",
                  "stats": {
                    "total": 3117,
                    "totsize": 5345361,
                    "os": "windows"
                  }
                }
              ]
            }


        The output is filtered by the :mod:`burpui.misc.acl` module so that you
        only see stats about the clients you are authorized to.

        :param server: Which server to collect data from when in multi-agent mode
        :type server: str

        :returns: The *JSON* described above
        """

        server = server or self.parser.parse_args()['serverName']
        self._check_acl(server)
        return self._get_clients_reports(server=server)

    def _check_acl(self, server):
        # Manage ACL
        if (not bui.standalone and not current_user.is_anonymous and
                (not current_user.acl.is_admin() and
                 not current_user.acl.is_server_allowed(server))):
            self.abort(403, 'Sorry, you don\'t have any rights on this server')

    def _get_clients_reports(self, res=None, server=None):
        args = self.parser.parse_args()
        limit = args['limit']
        aggregation = self.translation.get(args['aggregation'], 'number')
        ret = self._parse_clients_reports(res, server)
        backups = backups_orig = ret.get('backups', [])
        clients = clients_orig = ret.get('clients', [])
        aggregate = False
        if limit > 1:
            limit -= 1
            aggregate = True
        if aggregation == 'none':
            aggregate = False
            limit = 0
        # limit the number of elements to return so the graphs stay readable
        if len(backups) > limit and limit > 0:
            if aggregation == 'number':
                backups = (
                    sorted(backups, key=lambda x: x.get('number'), reverse=True)
                )[:limit]
            else:
                clients = (
                    sorted(clients, key=lambda x: x.get('stats', {}).get(aggregation), reverse=True)
                )[:limit]
        else:
            aggregate = False
        if aggregation == 'number':
            clients_name = [x.get('name') for x in backups]
            ret['backups'] = backups
            ret['clients'] = [
                x for x in clients_orig
                if x.get('name') in clients_name
            ]
        else:
            clients_name = [x.get('name') for x in clients]
            ret['clients'] = clients
            ret['backups'] = [
                x for x in backups_orig
                if x.get('name') in clients_name
            ]
        if aggregate:
            backups = {'name': 'others', 'number': 0}
            for client in backups_orig:
                if client.get('name') not in clients_name:
                    backups['number'] += client.get('number', 0)

            complement = {
                'name': 'others',
                'stats': {
                    'total': 0,
                    'totsize': 0,
                    'os': None
                }
            }
            # TODO: fix OS aggregation
            for client in clients_orig:
                if client.get('name') not in clients_name:
                    complement['stats']['total'] += client.get('stats', {}).get('total', 0)
                    complement['stats']['totsize'] += client.get('stats', {}).get('totsize', 0)
                    os = client.get('stats', {}).get('os', 'unknown')
                    if not complement['stats']['os']:
                        complement['stats']['os'] = os
                    elif os != complement['stats']['os']:
                        complement['stats']['os'] = 'unknown'

            ret['clients'].append(complement)
            ret['backups'].append(backups)

        return ret

    def _parse_clients_reports(self, res=None, server=None):
        if not res:
            try:
                clients = bui.client.get_all_clients(agent=server)
            except BUIserverException as e:
                self.abort(500, str(e))
            if not current_user.is_anonymous and not current_user.acl.is_admin():
                clients = [x for x in clients if current_user.acl.is_client_allowed(x['name'], server)]
            return bui.client.get_clients_report(clients, server)
        if bui.standalone:
            ret = res
        else:
            ret = res.get(server, {})
        if not current_user.is_anonymous and not current_user.acl.is_admin():
            ret['backups'] = [x for x in ret.get('backups', []) if current_user.acl.is_client_allowed(x.get('name'), server)]
            ret['clients'] = [x for x in ret.get('clients', []) if current_user.acl.is_client_allowed(x.get('name'), server)]
        return ret


@ns.route('/stats',
          '/<server>/stats',
          endpoint='clients_stats')
@ns.doc(
    params={
        'server': 'Which server to collect data from when in multi-agent mode',
    },
)
class ClientsStats(Resource):
    """The :class:`burpui.api.clients.ClientsStats` resource allows you to
    access general statistics about your clients.

    This resource is part of the :mod:`burpui.api.clients` module.

    An optional ``GET`` parameter called ``serverName`` is supported when running
    in multi-agent mode.
    """
    parser = ns.parser()
    parser.add_argument('serverName', help='Which server to collect data from when in multi-agent mode')
    client_fields = ns.model('ClientsStatsSingle', {
        'last': fields.DateTime(required=True, dt_format='iso8601', description='Date of last backup'),
        'name': fields.String(required=True, description='Client name'),
        'state': fields.LocalizedString(required=True, description='Current state of the client (idle, backup, etc.)'),
        'phase': fields.String(description='Phase of the current running backup'),
        'percent': fields.Integer(description='Percentage done', default=0),
        'labels': fields.List(fields.String, description='List of labels'),
    })

    @cache.cached(timeout=1800, key_prefix=cache_key, unless=force_refresh)
    @ns.marshal_list_with(client_fields, code=200, description='Success')
    @ns.expect(parser)
    @ns.doc(
        responses={
            403: 'Insufficient permissions',
            500: 'Internal failure',
        },
    )
    @browser_cache(1800)
    def get(self, server=None):
        """Returns a list of clients with their states

        **GET** method provided by the webservice.

        The *JSON* returned is:
        ::

            {
              [
                {
                  "last": "2015-05-17 11:40:02",
                  "name": "client1",
                  "state": "idle",
                  "phase": "phase1",
                  "percent": 12,
                  "labels": []
                },
                {
                  "last": "never",
                  "name": "client2",
                  "state": "idle",
                  "phase": "phase2",
                  "percent": 42,
                  "labels": []
                }
              ]
            }


        The output is filtered by the :mod:`burpui.misc.acl` module so that you
        only see stats about the clients you are authorized to.

        :param server: Which server to collect data from when in multi-agent mode
        :type server: str

        :returns: The *JSON* described above
        """

        server = server or self.parser.parse_args()['serverName']
        try:
            if (not bui.standalone and not current_user.is_anonymous and
                    (not current_user.acl.is_admin() and
                     not current_user.acl.is_server_allowed(server))):
                self.abort(403, 'Sorry, you don\'t have any rights on this server')
            jso = bui.client.get_all_clients(agent=server)
            if not current_user.is_anonymous and not current_user.acl.is_admin():
                jso = [x for x in jso if current_user.acl.is_client_allowed(x['name'], server)]
        except BUIserverException as e:
            self.abort(500, str(e))
        ret = []
        for client in jso:
            tmp_client = client
            try:
                labels = ClientLabels._get_labels(client['name'], server)
            except BUIserverException as exp:
                self.abort(500, str(exp))
            tmp_client['labels'] = labels
            ret.append(tmp_client)

        return ret


@ns.route('/all',
          '/<server>/all',
          endpoint='clients_all')
@ns.doc(
    params={
        'server': 'Which server to collect data from when in multi-agent mode',
    },
)
class AllClients(Resource):
    """The :class:`burpui.api.clients.AllClients` resource allows you to
    retrieve a list of all clients with their associated server (if any).

    An optional ``GET`` parameter called ``serverName`` is supported when
    running in multi-agent mode.
    """
    parser = ns.parser()
    parser.add_argument('serverName', help='Which server to collect data from when in multi-agent mode')
    parser.add_argument('user', help='For which user do we want the data (only works for admins')
    client_fields = ns.model('AllClients', {
        'name': fields.String(required=True, description='Client name'),
        'agent': fields.String(required=False, default=None, description='Associated Agent name'),
    })

    @cache.cached(timeout=1800, key_prefix=cache_key, unless=force_refresh)
    @ns.marshal_list_with(client_fields, code=200, description='Success')
    @ns.expect(parser)
    @ns.doc(
        responses={
            200: 'Success',
            403: 'Insufficient permissions',
        },
    )
    @browser_cache(1800)
    def get(self, server=None):
        """Returns a list of all clients with their associated Agent if any

        **GET** method provided by the webservice.

        The *JSON* returned is:

        ::

            [
              {
                "name": "client1",
                "agent": "agent1"
              },
              {
                "name": "client2",
                "agent": "agent1"
              },
              {
                "name": "client3",
                "agent": "agent2"
              }
            ]


        The output is filtered by the :mod:`burpui.misc.acl` module so that you
        only see stats about the clients you are authorized to.

        :param server: Which server to collect data from when in multi-agent mode
        :type server: str

        :returns: The *JSON* described above
        """
        ret = []
        is_admin = True
        args = self.parser.parse_args()
        server = server or args['serverName']

        is_admin = current_user.is_anonymous or current_user.acl.is_admin()
        is_moderator = current_user.is_anonymous or current_user.acl.is_moderator()

        user = (args.get('user', current_user.name) or current_user.name) if \
            is_admin or is_moderator \
            else current_user.name

        # drop privileges when switching user
        if user != current_user.name:
            is_admin = False
            is_moderator = False

        if (server and
                not is_admin and not is_moderator and
                not current_user.acl.is_server_allowed(server)):
            self.abort(403, "You are not allowed to view this server infos")

        if server:
            try:
                clients = [x['name'] for x in bui.client.get_all_clients(agent=server)]
            except BUIserverException:
                clients = []
            if not is_admin:
                # use the bui.acl module since we impersonalized the user
                ret = [{'name': x, 'agent': server} for x in clients if bui.acl.is_client_allowed(user, x, server)]
            else:
                ret = [{'name': x, 'agent': server} for x in clients]
            return ret

        if bui.standalone:
            try:
                clients = [x['name'] for x in bui.client.get_all_clients()]
            except BUIserverException:
                clients = []
            if not is_admin:
                ret = [{'name': x} for x in clients if bui.acl.is_client_allowed(user, x)]
            else:
                ret = [{'name': x} for x in clients]
        else:
            grants = {}
            clients_cache = {}
            for serv in bui.client.servers:
                try:
                    clients = [x['name'] for x in bui.client.get_all_clients(serv)]
                    clients_cache[serv] = clients
                except BUIserverException:
                    clients = []
            if not is_admin:
                for serv in bui.client.servers:
                    grants[serv] = [x for x in clients_cache.get(serv, []) if bui.acl.is_client_allowed(user, x, serv)]
            else:
                for serv in bui.client.servers:
                    grants[serv] = 'all'
            for (serv, clients) in iteritems(grants):
                if not isinstance(clients, list):
                    clients = clients_cache.get(serv, [])
                ret += [{'name': x, 'agent': serv} for x in clients]

        return ret



# -*- coding: utf8 -*-
"""
.. module:: burpui.api.custom.my_fields
    :platform: Unix
    :synopsis: Burp-UI api custom fields module.

.. moduleauthor:: Ziirish <hi+burpui@ziirish.me>


"""
import re
import datetime
import inspect
import arrow
import fnmatch

from ...ext.i18n import get_locale

from flask_restplus import fields
from flask_babel import gettext as _
from flask import escape
from tzlocal import get_localzone

TZ = str(get_localzone())


class DateTime(fields.DateTime):
    """Custom DateTime parser on top of :mod:`arrow` to provide lossless dates"""
    def parse(self, value):
        """Parse the value"""
        if value is None:
            return None
        try:
            value = float(value)
            a = arrow.get(datetime.datetime.utcfromtimestamp(value))
            a = a.replace(tzinfo=TZ)
        except (ValueError, TypeError):
            a = arrow.get(value)
            a = a.to(TZ)
        return fields.DateTime.parse(self, a.datetime)

    def format(self, value):
        """Format the value"""
        try:
            new_value = self.parse(value)
            if not new_value:
                return value
            if self.dt_format == 'iso8601':
                return new_value.isoformat()
            else:
                return fields.DateTime.format(self, value)
        except (AttributeError, ValueError) as e:
            raise fields.MarshallingError(e)
        except arrow.parser.ParserError as e:
            return _(str(value))


class DateTimeHuman(fields.Raw):
    """Custom parser to display human readable times (like '1 hour ago')"""
    def parse(self, value):
        """Parse the value"""
        if value is None:
            return None
        try:
            return arrow.get(value)
        except arrow.parser.ParserError:
            return None

    def format(self, value):
        """Format the value"""
        new_value = self.parse(value)
        if new_value:
            locale = get_locale()
            return new_value.humanize(locale=locale)
        return _(str(value))


class BackupNumber(fields.String):
    """Custom BackupNumber field"""
    def format(self, value):
        """Format the value"""
        try:
            return fields.String.format(self, '{0:07d}'.format(int(value)))
        except ValueError as e:
            raise fields.MarshallingError(e)


class SafeString(fields.String):
    """Custom SafeString field to encode HTML entities"""
    def format(self, value):
        """Format the value"""
        return fields.String.format(self, escape(value))


class LocalizedString(fields.String):
    """Custom LocalizedString to return localized strings"""
    def format(self, value):
        """Format the value"""
        return fields.String.format(self, _(value))


class Wildcard(fields.List):
    exclude = []
    _exclude = ['__dict__', '__doc__', '__module__', '__weakref__']
    # keep a track of the last object
    _idx = 0
    # cache the flat object
    _flat = None
    _obj = None
    _cache = []
    _last = None

    def _flatten(self, obj):
        if obj == self._obj and self._flat:
            return self._flat
        if isinstance(obj, dict):
            self._flat = obj.items()
        else:
            self._flat.extend(inspect.getmembers(obj, match_attributes))

        self._idx = 0
        self._cache = []
        self._obj = obj
        return self._flat

    @property
    def key(self):
        return self._last

    def output(self, key, obj):
        flat = self._flatten(obj)
        value = None
        reg = fnmatch.translate(key)

        for idx, (objkey, val) in enumerate(flat):
            if idx < self._idx:
                continue
            if objkey not in self._cache and \
                    objkey not in self.exclude and \
                    re.match(reg, objkey, re.IGNORECASE):
                value = val
                self._cache.append(objkey)
                self._last = objkey
                self._idx = idx
                break

        if value is None:
            if self.default is not None:
                return self.default
            return None

        return self.container.format(value)


def match_attributes(attr):
    return not(inspect.isroutine(attr) or
               (attr.__name__.startswith('__') and
                attr.__name__.endswith('__')))



# -*- coding: utf8 -*-
"""
.. module:: burpui.api.prefs
    :platform: Unix
    :synopsis: Burp-UI prefs api module.

.. moduleauthor:: Ziirish <hi+burpui@ziirish.me>

"""
from six import viewkeys
from flask import session, current_app, request
from flask_login import current_user
from werkzeug.datastructures import MultiDict

from . import api
from ..server import BUIServer  # noqa
from ..ext.i18n import LANGUAGES
from .custom import Resource

bui = current_app  # type: BUIServer
ns = api.namespace('preferences', 'Preferences methods')


@ns.route('/ui',
          endpoint='prefs_ui')
class PrefsUI(Resource):
    """The :class:`burpui.api.prefs.PrefsUI` resource allows you to
    set your UI preferences.

    This resource is part of the :mod:`burpui.api.prefs` module.
    """
    parser = ns.parser()
    parser.add_argument(
        'pageLength',
        type=int,
        required=False,
        help='Number of element per page'
    )
    parser.add_argument(
        'language',
        type=str,
        required=False,
        help='Language',
        choices=list(LANGUAGES.keys())
    )
    parser.add_argument(
        'dateFormat',
        type=str,
        required=False,
        help='Date format'
    )

    @staticmethod
    def _user_language(language):
        """Set the current user language"""
        if current_user and not current_user.is_anonymous and language:
            setattr(current_user, 'language', language)

    def _store_prefs(self, key, val):
        """Store the prefs if persistent storage is enabled"""
        if bui.config['WITH_SQL'] and not bui.config['BUI_DEMO']:
            from ..ext.sql import db
            from ..models import Pref
            pref = Pref.query.filter_by(user=current_user.name, key=key).first()
            if pref:
                if val:
                    pref.value = val
                else:
                    db.session.delete(pref)
            elif val:
                pref = Pref(current_user.name, key, val)
                db.session.add(pref)
            try:
                db.session.commit()
            except:
                db.session.rollback()

    def _update_prefs(self):
        """update prefs"""
        args = self.parser.parse_args()
        sess = session._get_current_object()
        ret = {}
        req = MultiDict()
        for loc in ['values', 'json']:
            data = getattr(request, loc, None)
            if data:
                req.update(data)
        for key in viewkeys(args):
            if key not in req:
                continue
            temp = args.get(key)
            if temp:
                if key == 'language':
                    self._user_language(temp)
                sess[key] = temp
            elif key in sess:
                del sess[key]
            ret[key] = temp
            self._store_prefs(key, temp)

        return ret

    @ns.doc(
        responses={
            200: 'Success',
            403: 'Insufficient permissions',
            404: 'No backend found',
        },
    )
    def get(self):
        """Returns a list of prefs

        **GET** method provided by the webservice.

        :returns: prefs
        """
        args = self.parser.parse_args()
        ret = {}
        sess = session
        for key in viewkeys(args):
            ret[key] = sess.get(key)
        return ret

    @ns.expect(parser, validate=True)
    @ns.doc(
        responses={
            201: 'Success',
            403: 'Not allowed',
            400: 'Missing parameters',
            404: 'Backend not found',
            500: 'Backend does not support this operation',
        },
    )
    def put(self):
        """Create prefs"""
        return self._update_prefs(), 201

    @ns.expect(parser)
    @ns.doc(
        responses={
            200: 'Success',
            403: 'Not allowed',
            400: 'Missing parameters',
            404: 'Backend not found',
            500: 'Backend does not support this operation',
        },
    )
    def delete(self):
        """Delete prefs"""
        args = self.parser.parse_args()
        sess = session
        ret = {}
        for key in viewkeys(args):
            temp = args.get(key)
            if temp:
                del sess[key]
                if bui.config['WITH_SQL']:
                    from ..ext.sql import db
                    from ..models import Pref
                    try:
                        Pref.query.filter_by(
                            user=current_user.name,
                            key=key
                        ).delete()
                        db.session.commit()
                    except:
                        db.session.rollback()
            ret[key] = sess.get(key)

        return ret

    @ns.expect(parser, validate=True)
    @ns.doc(
        responses={
            200: 'Success',
            403: 'Not allowed',
            400: 'Missing parameters',
            404: 'Backend not found',
            500: 'Backend does not support this operation',
        },
    )
    def post(self):
        """Change prefs"""
        return self._update_prefs()



# -*- coding: utf8 -*-
"""
Burp-UI is a web-ui for burp backup written in python with Flask and
jQuery/Bootstrap

.. module:: burpui.app
    :platform: Unix
    :synopsis: Burp-UI app module.

.. moduleauthor:: Ziirish <hi+burpui@ziirish.me>
"""
import os
import time
import logging

from logging import Formatter

from .desc import __version__, __release__
from .extensions import create_celery, create_db, create_websocket, \
    parse_db_setting, get_redis_server


def create_app(conf=None, verbose=0, logfile=None, **kwargs):
    """Initialize the whole application.

    :param conf: Configuration file to use
    :type conf: str

    :param verbose: Set the verbosity level
    :type verbose: int

    :param logfile: Store the logs in the given file
    :type logfile: str

    :param kwargs: Extra options:
                   - gunicorn (bool): Enable gunicorn engine instead of flask's
                   default. Default is True.
                   - unittest (bool): Are we running tests (used for test only).
                   Default is False.
                   - debug (bool): Enable debug mode. Default is False.
                   - cli (bool): Are we running the CLI. Default is False.
                   - reverse_proxy (bool): Are we behind a reverse-proxy.
                   Default is True if gunicorn is True
    :type kwargs: dict

    :returns: A :class:`burpui.server.BUIServer` object
    """
    from flask import g, request, session, render_template
    from flask_login import LoginManager
    from flask_bower import Bower
    from flask_babel import gettext
    from .utils import ReverseProxied, lookup_file, is_uuid
    from .security import basic_login_from_request
    from .server import BUIServer as BurpUI
    from .sessions import session_manager
    from .ext.cache import cache
    from .ext.i18n import babel, get_locale
    from .misc.auth.handler import BUIanon

    logger = logging.getLogger('burp-ui')

    gunicorn = kwargs.get('gunicorn', True)
    unittest = kwargs.get('unittest', False)
    debug = kwargs.get('debug', False)
    cli = kwargs.get('cli', False)
    reverse_proxy = kwargs.get('reverse_proxy', gunicorn)
    celery_worker = kwargs.get('celery_worker', False)
    websocket_server = kwargs.get('websocket_server', False)

    # The debug argument used to be a boolean so we keep supporting this format
    if isinstance(verbose, bool):
        if verbose:
            verbose = logging.DEBUG
        else:
            verbose = logging.CRITICAL
    else:
        levels = [
            logging.CRITICAL,
            logging.ERROR,
            logging.WARNING,
            logging.INFO,
            logging.DEBUG
        ]
        if verbose >= len(levels):
            verbose = len(levels) - 1
        if not verbose:
            verbose = 0
        verbose = levels[verbose]

    if logfile:
        from logging.handlers import RotatingFileHandler
        handler = RotatingFileHandler(
            logfile,
            maxBytes=1024 * 1024 * 100,
            backupCount=5
        )
    else:
        from logging import StreamHandler
        handler = StreamHandler()

    if verbose > logging.DEBUG:
        LOG_FORMAT = (
            '[%(asctime)s] %(levelname)s in '
            '%(module)s.%(funcName)s: %(message)s'
        )
    else:
        LOG_FORMAT = (
            '-' * 27 +
            '[%(asctime)s]' +
            '-' * 28 + '\n' +
            '%(levelname)s in %(module)s.%(funcName)s ' +
            '[%(pathname)s:%(lineno)d]:\n' +
            '%(message)s\n' +
            '-' * 80
        )

    handler.setLevel(verbose)
    handler.setFormatter(Formatter(LOG_FORMAT))

    logger.setLevel(verbose)

    logger.addHandler(handler)

    logger.debug(
        'conf: {}\n'.format(conf) +
        'verbose: {}\n'.format(logging.getLevelName(verbose)) +
        'logfile: {}\n'.format(logfile) +
        'gunicorn: {}\n'.format(gunicorn) +
        'debug: {}\n'.format(debug) +
        'unittest: {}\n'.format(unittest) +
        'cli: {}\n'.format(cli) +
        'reverse_proxy: {}'.format(reverse_proxy)
    )

    if not unittest:  # pragma: no cover
        from ._compat import patch_json
        patch_json()

    # We initialize the core
    app = BurpUI()
    if verbose:
        app.enable_logger()
    app.gunicorn = gunicorn

    app.config['CFG'] = None

    # Some config
    app.config['BUI_CLI'] = cli

    # FIXME: strange behavior when bundling errors
    # app.config['BUNDLE_ERRORS'] = True

    app.config['REMEMBER_COOKIE_HTTPONLY'] = True

    if debug and not gunicorn:  # pragma: no cover
        app.config['DEBUG'] = True and not unittest
        app.config['TESTING'] = True and not unittest

    # Still need to test conf file here because the init function can be called
    # by gunicorn directly
    if conf:
        app.config['CFG'] = lookup_file(conf, guess=False)
    else:
        app.config['CFG'] = lookup_file()

    logger.info('Using configuration: {}'.format(app.config['CFG']))

    app.setup(app.config['CFG'], unittest, cli)

    if debug:
        app.config.setdefault('TEMPLATES_AUTO_RELOAD', True)
        app.config['TEMPLATES_AUTO_RELOAD'] = True
        app.config['DEBUG'] = True

    SENTRY_AVAILABLE = False
    if app.demo:
        try:
            from .ext.sentry import sentry
            sentry.init_app(app, dsn=app.config['BUI_DSN'])
            SENTRY_AVAILABLE = True
        except ImportError:
            pass

    # manage application secret key
    if app.secret_key and \
            (app.secret_key.lower() == 'none' or
             (app.secret_key.lower() == 'random' and
              gunicorn)):  # pragma: no cover
        logger.critical('Your setup is not secure! Please consider setting a'
                        ' secret key in your configuration file')
        app.secret_key = 'Burp-UI'
    if not app.secret_key or app.secret_key.lower() == 'random':
        from base64 import b64encode
        app.secret_key = b64encode(os.urandom(256))

    app.wsgi_app = ReverseProxied(app.wsgi_app, app)

    # Manage reverse_proxy special tricks & improvements
    if reverse_proxy:  # pragma: no cover
        from werkzeug.contrib.fixers import ProxyFix

        app.wsgi_app = ProxyFix(app.wsgi_app)

    if app.storage and app.storage.lower() == 'redis':
        try:
            # Session setup
            if not app.session_db or \
                    str(app.session_db).lower() not in ['none', 'false']:
                from redis import Redis
                from .ext.session import sess
                host, port, pwd = get_redis_server(app)
                db = 0
                if app.session_db and \
                        str(app.session_db).lower() not \
                        in ['redis', 'default', 'true']:
                    try:  # pragma: no cover
                        (_, _, pwd, host, port, db) = \
                            parse_db_setting(app.session_db)
                    except ValueError as exp:
                        logger.warning(str(exp))
                try:
                    db = int(db)
                except ValueError:
                    db = 0
                logger.debug(
                    'SESSION: Using redis://guest:****@{}:{}/{}'.format(
                        host,
                        port,
                        db)
                )
                red = Redis(host=host, port=port, db=db, password=pwd)
                app.config['WITH_SRV_SESSION'] = True
                app.config['SESSION_TYPE'] = 'redis'
                app.config['SESSION_REDIS'] = red
                app.config['SESSION_USE_SIGNER'] = app.secret_key is not None
                app.config['SESSION_PERMANENT'] = False
                sess.init_app(app)
                session_manager.backend = red
        except Exception as exp:  # pragma: no cover
            logger.warning('Unable to initialize session: {}'.format(str(exp)))
            app.config['WITH_SRV_SESSION'] = False
        try:
            # Cache setup
            if not app.cache_db or \
                    str(app.cache_db).lower() not in ['none', 'false']:
                host, port, pwd = get_redis_server(app)
                db = 1
                if app.cache_db and \
                        str(app.cache_db).lower() not \
                        in ['redis', 'default', 'true']:
                    try:  # pragma: no cover
                        (_, _, pwd, host, port, db) = \
                            parse_db_setting(app.cache_db)
                    except ValueError as exp:
                        logger.warning(str(exp))
                try:
                    db = int(db)
                except ValueError:
                    db = 1
                logger.debug('CACHE: Using redis://guest:****@{}:{}/{}'.format(
                    host,
                    port,
                    db)
                )
                cache.init_app(
                    app,
                    config={
                        'CACHE_TYPE': 'redis',
                        'CACHE_REDIS_HOST': host,
                        'CACHE_REDIS_PORT': port,
                        'CACHE_REDIS_PASSWORD': pwd,
                        'CACHE_REDIS_DB': db
                    }
                )
                # clear cache at startup in case we removed or added servers
                with app.app_context():
                    cache.clear()
            else:  # pragma: no cover
                cache.init_app(app)
        except Exception as exp:  # pragma: no cover
            logger.warning('Unable to initialize cache: {}'.format(str(exp)))
            cache.init_app(app)
        try:
            # Limiter setup
            if app.limiter and str(app.limiter).lower() not \
                    in ['none', 'false']:  # pragma: no cover
                from .ext.limit import limiter
                app.config['RATELIMIT_HEADERS_ENABLED'] = True
                if app.limiter and str(app.limiter).lower() not \
                        in ['default', 'redis', 'true']:
                    app.config['RATELIMIT_STORAGE_URL'] = app.limiter
                else:
                    db = 3
                    host, port, pwd = get_redis_server(app)
                    if pwd:
                        conn = 'redis://guest:{}@{}:{}/{}'.format(
                            pwd,
                            host,
                            port,
                            db
                        )
                    else:
                        conn = 'redis://{}:{}/{}'.format(host, port, db)
                    app.config['RATELIMIT_STORAGE_URL'] = conn

                (_, _, pwd, host, port, db) = parse_db_setting(
                    app.config['RATELIMIT_STORAGE_URL']
                )

                logger.debug(
                    'LIMITER: Using redis://guest:****@{}:{}/{}'.format(
                        host,
                        port,
                        db
                    )
                )
                limiter.init_app(app)
                app.config['WITH_LIMIT'] = True
        except ImportError:  # pragma: no cover
            logger.warning('Unable to load limiter. Did you run \'pip install '
                           'flask-limiter\'?')
        except Exception as exp:  # pragma: no cover
            logger.warning('Unable to initialize limiter: {}'.format(str(exp)))
    else:
        cache.init_app(app)

    # Initialize i18n
    babel.init_app(app)

    # Create SQLAlchemy if enabled
    create_db(app, cli, unittest, celery_worker=celery_worker)

    if not celery_worker:
        from .api import api, apibp
        from .routes import view, mypad

        app.jinja_env.globals.update(
            isinstance=isinstance,
            list=list,
            mypad=mypad,
            version_id='{}-{}'.format(__version__, __release__),
        )

        # We initialize the API
        api.load_all()
        app.register_blueprint(apibp)

        # Then we load our routes
        app.register_blueprint(view)

        # Initialize Bower ext
        app.config.setdefault(
            'BOWER_COMPONENTS_ROOT',
            os.path.join('static', 'vendor')
        )
        app.config.setdefault('BOWER_REPLACE_URL_FOR', True)
        bower = Bower()
        bower.init_app(app)

    # Order of the initialization matters!
    # The websocket must be configured prior to the celery worker for instance

    # Initialize Session Manager
    session_manager.init_app(app)

    # And the login_manager
    app.login_manager = LoginManager()
    app.login_manager.anonymous_user = BUIanon
    app.login_manager.login_view = 'view.login'
    app.login_manager.login_message_category = 'info'
    app.login_manager.session_protection = 'strong'
    # This is just to have the strings in the .po files
    app.login_manager.login_message = gettext(
        'Please log in to access this page.'
    )
    app.login_manager.needs_refresh_message = gettext(
        'Please reauthenticate to access this page.'
    )
    # This will be called at runtime and will then translate the strings
    app.login_manager.localize_callback = gettext
    app.login_manager.init_app(app)

    # Create WebSocket server
    create_websocket(app, websocket_server, celery_worker, gunicorn, cli)

    # Create celery app if enabled
    create_celery(app, warn=False)

    def _check_session(user, request, api=False):
        """Check if the session is in the db"""
        if user and not session_manager.session_in_db():  # pragma: no cover
            login = getattr(user, 'name', None)
            if login and not is_uuid(login):
                remember = session.get('persistent', False)
                if not remember:
                    from flask_login import decode_cookie
                    remember_cookie = request.cookies.get(
                        app.config.get('REMEMBER_COOKIE_NAME'),
                        False
                    )
                    # check if the remember_cookie is legit
                    if remember_cookie and decode_cookie(remember_cookie):
                        remember = True
                session_manager.store_session(
                    login,
                    request.remote_addr,
                    request.headers.get('User-Agent'),
                    remember,
                    api
                )
            elif login:
                app.uhandler.remove(login)

    @app.before_request
    def setup_request():
        g.version = '{}-{}'.format(__version__, __release__)
        g.locale = get_locale()
        g.now = round(time.time())
        g.date_format = session.get('dateFormat', 'llll')
        # make sure to store secure cookie if required
        if app.scookie:
            criteria = [
                request.is_secure,
                request.headers.get('X-Forwarded-Proto', 'http') == 'https'
            ]
            app.config['SESSION_COOKIE_SECURE'] = \
                app.config['REMEMBER_COOKIE_SECURE'] = any(criteria)
        if '_extra' in request.args:
            session['_extra'] = request.args.get('_extra')
        g._extra = session.get('_extra', '')

    @app.login_manager.user_loader
    def load_user(userid):
        """User loader callback"""
        if app.auth != 'none':
            user = app.uhandler.user(userid)
            if not user:
                return None
            if 'X-Language' in request.headers:
                language = request.headers.get('X-Language')
                user.language = language
                session['language'] = language
            if '_id' not in session:
                from flask_login import login_user
                # if _id is not in session, it means we loaded the user from
                # cache/db using the remember cookie so we need to login it
                login_user(user, remember=user.is_authenticated, fresh=False)
            _check_session(user, request)
            return user
        return None

    @app.login_manager.request_loader
    def load_user_from_request(request):
        """User loader from request callback"""
        if app.auth != 'none':
            user = basic_login_from_request(request, app)
            _check_session(user, request, True)
            return user

    @app.after_request
    def after_request(response):
        if getattr(g, 'basic_session', False):
            if session_manager.invalidate_current_session():
                session_manager.delete_session()
        return response

    return app

    if app.demo and SENTRY_AVAILABLE:
        @app.errorhandler(500)
        def internal_server_error(error):
            from .ext.sentry import sentry
            return render_template(
                '500_sentry.html',
                event_id=g.sentry_event_id,
                public_dsn=sentry.client.get_public_dsn('https')
            )


init = create_app



# -*- coding: utf8 -*-
"""
.. module:: burpui.ws.namespace
    :platform: Unix
    :synopsis: Burp-UI WebSocket namespace module.

.. moduleauthor:: Ziirish <hi+burpui@ziirish.me>

"""
from flask import current_app, request
from flask_socketio import Namespace, emit

from ..server import BUIServer  # noqa

bui = current_app  # type: BUIServer


class BUINamespace(Namespace):
    def on_connect(self):
        sid = request.sid
        # if current_user.is_authenticated:
        bui.logger.debug('Someone just connected! {}'.format(sid))
        # else:
        #    bui.logger.debug('Illegal connection')
        #    disconnect()
        #    return False

    def on_disconnect(self):
        sid = request.sid
        bui.logger.debug('Someone just disconnected! {}'.format(sid))

    def on_echo(self, data):
        emit('reply', data, namespace='/ws')
        bui.logger.debug('Got message: {}'.format(data))



# -*- coding: utf8 -*-
"""
.. module:: burpui.security
    :platform: Unix
    :synopsis: Burp-UI security module.

.. moduleauthor:: Ziirish <hi+burpui@ziirish.me>

"""
from ._compat import to_unicode, urlparse, urljoin


def sanitize_string(string, strict=True, paranoid=False):
    """Return a 'safe' version of the string (ie. remove malicious chars like
    '\n')

    :param string: String to escape
    :type string: str
    """
    if not string:
        return ""
    if paranoid:
        return to_unicode(string.encode('unicode_escape'))
    elif strict:
        return to_unicode(string).split('\n')[0]
    else:
        import re
        ret = repr(string).replace('\\\\', '\\')
        ret = re.sub(r"^u?(?P<quote>['\"])(.*)(?P=quote)$", r"\2", ret)
        return to_unicode(ret)


def basic_login_from_request(request, app):
    """Check 'Authorization' headers and log the user in if possible.

    :param request: The input request
    :type request: :class:`flask.Request`

    :param app: The application context
    :type app: :class:`burpui.server.BUIServer`
    """
    if app.auth != 'none':
        if request.headers.get('X-From-UI', False):
            return None
        auth = request.authorization
        if auth:
            from flask import session, g
            app.logger.debug('Found Basic user: {}'.format(auth.username))
            refresh = False
            if 'login' in session and session['login'] != auth.username:
                refresh = True
                session.clear()
                session['login'] = auth.username
            session['language'] = request.headers.get('X-Language', 'en')
            user = app.uhandler.user(auth.username, refresh)
            if user.active and user.login(auth.password):
                from flask_login import login_user
                from .sessions import session_manager
                login_user(user)
                if request.headers.get('X-Reuse-Session', False):
                    session_manager.store_session(
                        auth.username,
                        request.remote_addr,
                        request.headers.get('User-Agent'),
                        remember=False,
                        api=True
                    )
                else:
                    g.basic_session = True
                app.logger.debug('Successfully logged in')
                return user
            app.logger.warning('Failed to log-in')
    return None


def is_safe_url(target):
    from flask import request
    ref_url = urlparse(request.host_url)
    test_url = urlparse(urljoin(request.host_url, target))
    return test_url.scheme in ('http', 'https') and \
        ref_url.netloc == test_url.netloc



# -*- coding: utf8 -*-
"""
.. module:: burpui.misc.parser.burp2
    :platform: Unix
    :synopsis: Burp-UI configuration file parser for Burp2.
.. moduleauthor:: Ziirish <hi+burpui@ziirish.me>
"""
from .burp1 import Parser as Burp1


def __(string):
    """dummy function to fake the translation"""
    return string


# inherit Burp1 parser so we can just override available options
class Parser(Burp1):
    """Extends :class:`burpui.misc.parser.burp1.Parser`"""
    pver = 2

    pair_srv = [
        u'port',
        u'max_children',
        u'status_port',
        u'max_status_children',
    ]
    pair_associations = {
        u'port': u'max_children',
        u'max_children': u'port',
        u'status_port': u'max_status_children',
        u'max_status_children': u'status_port',
    }
    integer_srv = Burp1.integer_srv
    for rem in ['port', 'max_children', 'status_port', 'max_status_children']:
        integer_srv.remove(rem)
    advanced_type = Burp1.advanced_type
    advanced_type.update({
        u'port': u'integer',
        u'max_children': u'integer',
        u'status_port': u'integer',
        u'max_status_children': u'integer',
    })
    multi_srv = Burp1.multi_srv + [
        u'label',
    ]
    string_srv = Burp1.string_srv + [
        u'manual_delete',
        u'rblk_memory_max',
    ]
    boolean_add = [
        u'acl',
        u'xattr',
        u'glob_after_script_pre',
        u'cname_fqdn',
        u'cname_lowercase',
    ]
    boolean_add_cli = [
        u'enabled',
    ]
    boolean_srv = Burp1.boolean_srv + boolean_add
    boolean_cli = Burp1.boolean_cli + boolean_add + boolean_add_cli
    multi_cli = Burp1.multi_cli + [
        u'label',
    ]
    integer_cli = Burp1.integer_cli + [
        u'randomise',
    ]
    fields_cli = Burp1.fields_cli + boolean_add + boolean_add_cli + [
        u'randomise',
        u'manual_delete',
        u'label',
    ]
    placeholders = Burp1.placeholders
    placeholders.update({
        u'acl': u"0|1",
        u'xattr': u"0|1",
        u'randomise': __(u"max secs"),
        u'manual_delete': __(u"path"),
        u'label': __(u"some informations"),
        u'status_address': __(u"address|localhost"),
        u'glob_after_script_pre': u"0|1",
        u'enabled': u"0|1",
        u'cname_fqdn': u"0|1",
        u'cname_lowercase': u"0|1",
        u'rblk_memory_max': u"b/Kb/Mb/Gb",
    })
    values = Burp1.values
    # status_address can now listen on any address
    del values['status_address']
    defaults = Burp1.defaults
    defaults.update({
        u'acl': True,
        u'xattr': True,
        u'glob_after_script_pre': True,
        u'randomise': 0,
        u'manual_delete': u'',
        u'label': u'',
        u'enabled': True,
        u'cname_fqdn': True,
        u'cname_lowercase': False,
        u'rblk_memory_max': u'256Mb',
    })
    doc = Burp1.doc
    doc.update({
        u'acl': __(u"If acl support is compiled into burp, this allows you to"
                   " decide whether or not to backup acls at runtime. The"
                   " default is '1'."),
        u'xattr': __(u"If xattr support is compiled into burp, this allows you"
                     " to decide whether or not to backup xattrs at runtime."
                     " The default is '1'."),
        u'randomise': __(u"When running a timed backup, sleep for a random"
                         " number of seconds (between 0 and the number given)"
                         " before contacting the server. Alternatively, this"
                         " can be specified by the '-q' command line option."),
        u'manual_delete': __(u"This can be overridden by the clientconfdir"
                             " configuration files in clientconfdir on the"
                             " server. When the server needs to delete old"
                             " backups, or rubble left over from generating"
                             " reverse patches with librsync=1, it will"
                             " normally delete them in place. If you use the"
                             " 'manual_delete' option, the files will be moved"
                             " to the path specified for deletion at a later"
                             " point. You will then need to configure a cron"
                             " job, or similar, to delete the files yourself."
                             " Do not specify a path that is not on the same"
                             " filesystem as the client storage directory."),
        u'label': __(u"You can have multiple labels, and they can be"
                     " overridden in the client configuration files in"
                     " clientconfdir on the server. They will appear as an"
                     " array of strings in the server status monitor JSON"
                     " output. The idea is to provide a mechanism for"
                     " arbitrary values to be passed to clients of the server"
                     " status monitor."),
        u'status_address': __(u"Defines the main TCP address that the server "
                              "listens on for status requests. The default  "
                              "is  special  value  'localhost'  that includes "
                              "both '::1' (if available) and '127.0.0.1' "
                              "(always)."),
        u'glob_after_script_pre': __(u"Set this to 0 if you do not want"
                                     " include_glob settings to be evaluated"
                                     " after the pre script is run. The"
                                     " default is 1."),
        u'enabled': __(u"Set this to 0 if you want to disable all clients. The"
                       " default is 1. This option can be overridden"
                       " per-client in the client configuration files in"
                       " clientconfdir on the server."),
        u'cname_fqdn': __(u"Whether to keep fqdn cname (like"
                          " 'testclient.example.com') when looking-up in"
                          " clientconfdir. This also affects the fqdn lookup"
                          " on the client (see client configuration options"
                          " for details). The default is 1. When set to 0, the"
                          " fqdn provided by the client while authenticating"
                          " will be stripped ('testclient.example.com'"
                          " becomes 'testclient')."),
        u'cname_lowercase': __(u"Whether to force lowercase cname when"
                               " looking-up in clientconfdir. This also"
                               " affects the fqdn lookup on the client (see"
                               " client configuration options for details)."
                               " The default is 0. When set to 1 the name"
                               " provided by the client while authenticating"
                               " will be lowercased."),
        u'port': __(u"Defines the main TCP port that the server listens on. "
                    "Specify multiple 'port' entries on separate lines in "
                    "order to listen on multiple ports. Each port can be "
                    "configured with its own 'max_children' value."),
        u'max_children': __(u"Defines the number of child processes to fork "
                            "(the number of clients that can simultaneously "
                            "connect. The default is 5. Specify multiple "
                            "'max_children' entries on separate lines if you "
                            "have configured multiple port entries."),
        u'status_port': __(u"Defines the TCP port that the server listens on "
                           "for status requests. Comment this out to have no "
                           "status server. Specify multiple 'status_port' "
                           "entries on separate lines in order to listen on "
                           "multiple ports. Each port can be configured with "
                           "its own 'max_status_children' value."),
        u'max_status_children': __(u"Defines the number of status child "
                                   "processes to fork (the number of status "
                                   "clients that can simultaneously connect. "
                                   "The default is 5. Specify multiple "
                                   "'max_status_children' entries on separate "
                                   "lines if you have configured multiple "
                                   "status_port entries."),
        u'rblk_memory_max': __("The maximum amount of data from the disk "
                               "cached in server memory during a protocol2 "
                               "restore/verify. The default is 256Mb. This "
                               "option can be overriden per-client in the "
                               "client configuration files in clientconfdir "
                               "on the server."),
    })



# -*- coding: utf8 -*-
import os

from .interface import BUIacl, BUIaclLoader
from .meta import meta_grants

from importlib import import_module
from six import iteritems
from collections import OrderedDict


class ACLloader(BUIaclLoader):
    section = name = 'ACL'

    def __init__(self, app=None):
        """See :func:`burpui.misc.acl.interface.BUIaclLoader.__init__`

        :param app: Application context
        :type app: :class:`burpui.server.BUIServer`
        """
        self.app = app
        self.conf = self.app.conf
        self._acl = ACLhandler(self)
        backends = []
        self.errors = {}
        if self.section in self.conf.options:
            opts = {}
            opts['extended'] = self.conf.safe_get(
                'extended',
                'boolean',
                section=self.section,
                defaults=True
            )
            opts['assume_rw'] = self.conf.safe_get(
                'assume_rw',
                'boolean',
                section=self.section,
                defaults=True
            )
            opts['legacy'] = self.conf.safe_get(
                'legacy',
                'boolean',
                section=self.section
            )
            opts['implicit_link'] = self.conf.safe_get(
                'implicit_link',
                'boolean',
                section=self.section,
                defaults=True
            )
            opts['inverse_inheritance'] = self.conf.safe_get(
                'inverse_inheritance',
                'boolean',
                section=self.section
            )
            meta_grants.options = opts
            meta_grants.init_app(app)
        if self.app.acl_engine and 'none' not in self.app.acl_engine:
            me, _ = os.path.splitext(os.path.basename(__file__))
            back = self.app.acl_engine
            for au in back:
                if au == me:
                    self.app.logger.critical('Recursive import not permitted!')
                    continue
                try:
                    (modpath, _) = __name__.rsplit('.', 1)
                    mod = import_module('.' + au, modpath)
                    obj = mod.ACLloader(self.app)
                    backends.append(obj)
                except:
                    import traceback
                    self.errors[au] = traceback.format_exc()
        for name, plugin in iteritems(self.app.plugin_manager.get_plugins_by_type('acl')):
            try:
                obj = plugin.ACLloader(self.app)
                backends.append(obj)
            except:
                import traceback
                self.errors[name] = traceback.format_exc()
        backends.sort(key=lambda x: getattr(x, 'priority', -1), reverse=True)
        if not backends:
            raise ImportError(
                'No backend found for \'{}\':\n{}'.format(self.app.acl_engine,
                                                          self.errors)
            )
        for name, err in iteritems(self.errors):
            self.app.logger.error(
                'Unable to load module {}:\n{}'.format(repr(name), err)
            )
        self.backends = OrderedDict()
        for obj in backends:
            self.backends[obj.name] = obj

    def reload(self):
        return None

    @property
    def acl(self):
        return self._acl

    @property
    def grants(self):
        return meta_grants.grants

    @property
    def groups(self):
        return meta_grants.groups


class ACLhandler(BUIacl):
    """See :class:`burpui.misc.acl.interface.BUIacl`"""
    def __init__(self, loader=None):
        """:func:`burpui.misc.acl.interface.BUIacl.__init__` instanciate ACL
        engine.

        :param loader: ACL loader
        :type loader: :class:`burpui.misc.acl.handler.ACLloader`
        """
        self.loader = loader

    def _iterate_through_loader(self, method, *args, **kwargs):
        ret = None
        for _, acl_engine in iteritems(self.loader.backends):
            func = getattr(acl_engine.acl, method)
            ret = func(*args, **kwargs)
            if isinstance(ret, tuple):
                (ret, _) = ret
            if ret:
                break
        if not ret:
            func = getattr(meta_grants, method)
            ret = func(*args, **kwargs)
            if isinstance(ret, tuple):
                (ret, _) = ret
        return ret

    def is_admin(self, username=None):
        """See :func:`burpui.misc.acl.interface.BUIacl.is_admin`"""
        ret = self._iterate_through_loader('is_admin', username) or False
        return ret

    def is_moderator(self, username=None):
        """See :func:`burpui.misc.acl.interface.BUIacl.is_moderator`"""
        ret = self._iterate_through_loader('is_moderator', username) or False
        return ret

    def is_client_rw(self, username=None, client=None, server=None):
        """See :func:`burpui.misc.acl.interface.BUIacl.is_client_rw`"""
        ret = self._iterate_through_loader(
            'is_client_rw',
            username,
            client,
            server
        ) or False
        return ret

    def is_client_allowed(self, username=None, client=None, server=None):
        """See :func:`burpui.misc.acl.interface.BUIacl.is_client_allowed`"""
        ret = self._iterate_through_loader(
            'is_client_allowed',
            username,
            client,
            server
        ) or False
        return ret

    def is_server_rw(self, username=None, server=None):
        """See :func:`burpui.misc.acl.interface.BUIacl.is_server_rw`"""
        ret = self._iterate_through_loader(
            'is_server_rw',
            username,
            server
        ) or False
        return ret

    def is_server_allowed(self, username=None, server=None):
        """See :func:`burpui.misc.acl.interface.BUIacl.is_server_allowed`"""
        ret = self._iterate_through_loader(
            'is_server_allowed',
            username,
            server
        ) or False
        return ret



# -*- coding: utf8 -*-
"""
Burp-UI is a web-ui for burp backup written in python with Flask and
jQuery/Bootstrap

.. module:: burpui.desc
    :platform: Unix
    :synopsis: Burp-UI desc module.

.. moduleauthor:: Ziirish <hi+burpui@ziirish.me>
"""
import os

__title__ = 'burp-ui'
__author__ = 'Benjamin SANS (Ziirish)'
__author_email__ = 'hi+burpui@ziirish.me'
__url__ = 'https://git.ziirish.me/ziirish/burp-ui'
__doc__ = 'https://burp-ui.readthedocs.io/en/latest/'
__description__ = ('Burp-UI is a web-ui for burp backup written in python with '
                   'Flask and jQuery/Bootstrap')
__license__ = 'BSD 3-clause'
__version__ = open(
    os.path.join(os.path.dirname(os.path.realpath(__file__)), 'VERSION')
).read().rstrip()
try:  # pragma: no cover
    __release__ = open(
        os.path.join(os.path.dirname(os.path.realpath(__file__)), 'RELEASE')
    ).read().rstrip()
except:  # pragma: no cover
    __release__ = 'unknown'



# -*- coding: utf8 -*-
"""
.. module:: burpui.ext.cache
    :platform: Unix
    :synopsis: Burp-UI external cache module.

.. moduleauthor:: Ziirish <hi+burpui@ziirish.me>

"""
from flask_caching import Cache


cache = Cache(config={
    'CACHE_TYPE': 'simple',
    'CACHE_THRESHOLD': 50,
    'CACHE_DEFAULT_TIMEOUT': 7200
})
